# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the workflow views."""

import re
from typing import Any, ClassVar, assert_never, override
from unittest import mock

from django.conf import settings
from django.db.models import Max
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.client.models import (
    Link,
    RuntimeParameter,
    RuntimeParameters,
    StrictBaseModel,
    WorkflowTemplateData,
    WorkflowTemplateDataNew,
)
from debusine.db.models import (
    Scope,
    WorkRequest,
    WorkflowTemplate,
    Workspace,
    default_workspace,
)
from debusine.db.playground import scenarios
from debusine.server.scopes import urlconf_scope
from debusine.server.serializers import WorkRequestSerializer
from debusine.server.views.tests.base import TestCase
from debusine.tasks import TaskConfigError
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestResponseType,
    override_permission,
)


class WorkflowTemplateViewSetTests(TestCase):
    """Tests for :py:class:`WorkflowTemplateViewSet`."""

    scenario = scenarios.DefaultContextAPI()
    sample: ClassVar[WorkflowTemplate]
    sample_client: ClassVar[WorkflowTemplateData]
    sample_client_with_links: ClassVar[WorkflowTemplateData]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sample = cls.playground.create_workflow_template(
            "sample", task_name="noop"
        )
        cls.sample_client = WorkflowTemplateData(
            id=cls.sample.id,
            name=cls.sample.name,
            task_name=cls.sample.task_name,
            priority=cls.sample.priority,
            static_parameters=cls.sample.static_parameters,
            runtime_parameters=cls.sample.runtime_parameters,
        )
        cls.sample_client_with_links = cls.sample_client.model_copy()
        cls.sample_client_with_links.links = cls.get_links(cls.sample)

    @classmethod
    def get_links(
        cls, workflow_template: WorkflowTemplate
    ) -> dict[str, str | Link]:
        return {
            "self": reverse(
                "api:workflow-template-detail",
                kwargs={
                    "workspace": workflow_template.workspace.name,
                    "pk": workflow_template.pk,
                },
            ),
            "webui_self": workflow_template.get_absolute_url(),
        }

    def _workspace_url_kwarg(self, workspace: Workspace | str | None) -> str:
        match workspace:
            case None:
                return self.scenario.workspace.name
            case str():
                return workspace
            case _:
                return workspace.name

    def _request_data(
        self, data: dict[str, Any] | StrictBaseModel | None = None
    ) -> dict[str, Any]:
        match data:
            case dict():
                return data
            case StrictBaseModel():
                return data.model_dump(mode="json")
            case None:
                return {}
            case _ as unreachable:
                assert_never(unreachable)

    def _reverse_list(
        self,
        workspace: Workspace | str | None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        kwargs["workspace"] = self._workspace_url_kwarg(workspace)
        return reverse("api:workflow-template-list", kwargs=kwargs)

    def _reverse_detail(
        self,
        workflow_template: int | str | WorkflowTemplate | None,
        workspace: Workspace | str | None,
        **kwargs: str,
    ) -> str:
        """Build the URL for a test request."""
        kwargs["workspace"] = self._workspace_url_kwarg(workspace)

        match workflow_template:
            case None:
                kwargs["pk"] = str(self.sample.pk)
            case int():
                kwargs["pk"] = str(workflow_template)
            case str():
                kwargs["pk"] = workflow_template
            case WorkflowTemplate():
                kwargs["pk"] = str(workflow_template.pk)
            case _ as unreachable:
                assert_never(unreachable)

        return reverse("api:workflow-template-detail", kwargs=kwargs)

    def _headers(self, user_token: bool) -> dict[str, str]:
        """Build headers for a test request."""
        match user_token:
            case True:
                return {"token": self.scenario.user_token.key}
            case False:
                return {}
            case _ as unreachable:
                assert_never(unreachable)

    def get(
        self,
        workflow_template: int | str | WorkflowTemplate | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a workflow template."""
        return self.client.get(
            self._reverse_detail(workflow_template, workspace),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def list(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for the workflow template list."""
        return self.client.get(
            self._reverse_list(workspace),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def post(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | StrictBaseModel | None = None,
    ) -> TestResponseType:
        """POST request."""
        return self.client.post(
            self._reverse_list(workspace),
            data=self._request_data(data),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def put(
        self,
        workflow_template: int | WorkflowTemplate | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | StrictBaseModel | None = None,
    ) -> TestResponseType:
        """PUT request for a workflow template."""
        return self.client.put(
            self._reverse_detail(workflow_template, workspace),
            data=self._request_data(data),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def patch(
        self,
        workflow_template: int | WorkflowTemplate | None = None,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | StrictBaseModel | None = None,
    ) -> TestResponseType:
        """PATCH request for a workflow template."""
        return self.client.patch(
            self._reverse_detail(workflow_template, workspace),
            data=self._request_data(data),
            content_type="application/json",
            headers=self._headers(user_token),
        )

    def assertSampleUnchanged(self) -> None:
        """Ensure self.sample is unchanged."""
        self.sample.refresh_from_db()
        self.assertEqual(self.sample.name, "sample")
        self.assertEqual(self.sample.task_name, "noop")
        self.assertEqual(self.sample.priority, 0)
        self.assertEqual(self.sample.static_parameters, {})
        self.assertEqual(self.sample.runtime_parameters, RuntimeParameter.ANY)

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        for method in ("get", "list", "post", "put", "patch"):
            with self.subTest(method=method):
                response = getattr(self, method)(user_token=False)
                self.assertResponseProblem(
                    response,
                    "Error",
                    detail_pattern=(
                        "Authentication credentials were not provided."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_workspace_not_found(self) -> None:
        """Workspace must exist."""
        for method in ("get", "list", "post", "put", "patch"):
            with self.subTest(method=method):
                response = getattr(self, method)(workspace="does-not-exist")
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace does-not-exist not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_workspace_not_accessible(self) -> None:
        """User must be able to display the workspace."""
        workspace = self.playground.create_workspace(name="private")
        for method in ("get", "list", "post", "put", "patch"):
            with self.subTest(method=method):
                response = getattr(self, method)(workspace=workspace)
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace private not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_not_found(self) -> None:
        """The workflow template must exist."""
        max_id = WorkflowTemplate.objects.aggregate(Max('id'))['id__max']
        for method in ("get", "put", "patch"):
            with self.subTest(method=method):
                response = getattr(self, method)(workflow_template=max_id + 1)
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_list_unauthorized(self) -> None:
        """The user must be able to display the collection."""
        with override_permission(WorkflowTemplate, "can_display", DenyAll):
            response = self.list()
            data = self.assertAPIResponseOk(response)
            self.assertEqual(
                data,
                {"count": 0, "next": None, "previous": None, "results": []},
            )

    def test_detail_unauthorized(self) -> None:
        """The user must be able to display the collection."""
        for method in ("get", "put", "patch"):
            with (
                self.subTest(method=method),
                override_permission(WorkflowTemplate, "can_display", DenyAll),
            ):
                response = getattr(self, method)()
                self.assertResponseProblem(
                    response,
                    "Object not found",
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_create_unauthorized(self) -> None:
        """The user must be able to create the collection."""
        with override_permission(Workspace, "can_configure", DenyAll):
            response = self.post(
                data={
                    "name": "test",
                    "task_name": "noop",
                    "static_parameters": {},
                    "priority": 0,
                }
            )
            self.assertResponseProblem(
                response,
                "playground cannot configure workspace"
                f" {self.scenario.workspace}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_create_unauthorized_create_child_workspace(self) -> None:
        """Only scope owners may create ``create_child_workspace`` templates."""
        with (
            override_permission(Workspace, "can_configure", AllowAll),
            override_permission(Scope, "can_create_workspace", DenyAll),
        ):
            response = self.post(
                data=WorkflowTemplateDataNew(
                    name="test",
                    task_name="create_child_workspace",
                    static_parameters={},
                )
            )
            self.assertResponseProblem(
                response,
                f"playground cannot create workspaces in {self.scenario.scope}",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_update_unauthorized(self) -> None:
        """The user must be able to update the collection."""
        for method in ("put", "patch"):
            with (
                self.subTest(method=method),
                override_permission(Workspace, "can_configure", DenyAll),
            ):
                response = getattr(self, method)(
                    data=WorkflowTemplateDataNew(
                        name="test", task_name="noop", static_parameters={}
                    )
                )
                self.assertResponseProblem(
                    response,
                    "playground cannot configure sample",
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_update_unauthorized_create_child_workspace(self) -> None:
        """Only scope owners may update ``create_child_workspace`` templates."""
        template = self.playground.create_workflow_template(
            name="test", task_name="create_child_workspace"
        )

        for method in ("put", "patch"):
            with (
                self.subTest(method=method),
                override_permission(
                    WorkflowTemplate, "can_configure", AllowAll
                ),
                override_permission(Scope, "can_create_workspace", DenyAll),
            ):
                response = getattr(self, method)(
                    workflow_template=template,
                    data=WorkflowTemplateDataNew(
                        name="test",
                        task_name="create_child_workspace",
                        static_parameters={},
                    ),
                )
                self.assertResponseProblem(
                    response,
                    "playground cannot create workspaces in"
                    f" {self.scenario.scope}",
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_post_unauthorized(self) -> None:
        """The user must be able to create the workflow template."""
        with override_permission(Workspace, "can_configure", DenyAll):
            response = self.post(
                data=WorkflowTemplateDataNew(
                    name="test", task_name="noop", static_parameters={}
                )
            )
            self.assertResponseProblem(
                response,
                "playground cannot configure workspace debusine/System",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_patch_unauthorized(self) -> None:
        """The user must be able to update the workflow template."""
        with override_permission(Workspace, "can_configure", DenyAll):
            response = self.patch(
                data=WorkflowTemplateDataNew(
                    name="test", task_name="noop", static_parameters={}
                )
            )
            self.assertResponseProblem(
                response,
                "playground cannot configure sample",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_get_success(self) -> None:
        response = self.get()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "id": self.sample.id,
                "name": self.sample.name,
                "task_name": self.sample.task_name,
                "static_parameters": self.sample.static_parameters,
                "runtime_parameters": self.sample.runtime_parameters,
                "priority": self.sample.priority,
                "links": self.get_links(self.sample),
                "extra_groups": [],
                "restricted": False,
            },
        )

    def test_get_name_success(self) -> None:
        response = self.get(workflow_template="sample")
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "id": self.sample.id,
                "name": self.sample.name,
                "task_name": self.sample.task_name,
                "static_parameters": self.sample.static_parameters,
                "runtime_parameters": self.sample.runtime_parameters,
                "priority": self.sample.priority,
                "links": self.get_links(self.sample),
                "extra_groups": [],
                "restricted": False,
            },
        )

    def test_get_honours_workspace(self) -> None:
        workspace = self.playground.create_workspace(
            name="workspace", public=True
        )
        template = WorkflowTemplate.objects.create(
            name=self.sample.name,
            workspace=workspace,
            task_name="noop",
            static_parameters={},
        )

        response = self.get(workspace=workspace, workflow_template=template)
        data = self.assertAPIResponseOk(response)
        self.assertEqual(data["id"], template.id)

        response = self.get(workflow_template=template)
        self.assertResponseProblem(
            response,
            title="Object not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_list_success(self) -> None:
        response = self.list()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "id": self.sample.pk,
                        "name": "sample",
                        "priority": 0,
                        "static_parameters": {},
                        "runtime_parameters": RuntimeParameter.ANY,
                        "task_name": "noop",
                        "links": self.get_links(self.sample),
                        "extra_groups": [],
                        "restricted": False,
                    }
                ],
            },
        )

    def test_list_different_workspace(self) -> None:
        workspace = self.playground.create_workspace(name="other", public=True)
        other2 = self.playground.create_workflow_template(
            task_name="noop", name="other2", workspace=workspace
        )
        other1 = self.playground.create_workflow_template(
            task_name="noop", name="other1", workspace=workspace
        )
        response = self.list(workspace=workspace)
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "count": 2,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "id": other1.pk,
                        "name": other1.name,
                        "priority": 0,
                        "static_parameters": {},
                        "runtime_parameters": RuntimeParameter.ANY,
                        "task_name": "noop",
                        "links": self.get_links(other1),
                        "extra_groups": [],
                        "restricted": False,
                    },
                    {
                        "id": other2.pk,
                        "name": other2.name,
                        "priority": 0,
                        "static_parameters": {},
                        "runtime_parameters": RuntimeParameter.ANY,
                        "task_name": "noop",
                        "links": self.get_links(other2),
                        "extra_groups": [],
                        "restricted": False,
                    },
                ],
            },
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_success(self) -> None:
        """Create a new workflow template."""
        static_parameters = {
            "target_distribution": "debian:bookworm",
            "architectures": ["amd64", "arm64"],
        }

        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="sbuild",
                static_parameters=static_parameters,
                runtime_parameters=RuntimeParameter.ANY,
            )
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=self.scenario.workspace
        )
        self.assertEqual(template.task_name, "sbuild")
        self.assertEqual(template.static_parameters, static_parameters)
        self.assertEqual(template.runtime_parameters, RuntimeParameter.ANY)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_duplicate_name(self) -> None:
        self.playground.create_workflow_template("duplicate", task_name="noop")

        with self.assert_model_count_unchanged(WorkflowTemplate):
            response = self.post(
                data=WorkflowTemplateDataNew(
                    name="duplicate",
                    task_name="noop",
                    static_parameters={},
                )
            )

        self.assertResponseProblem(
            response,
            "Workflow template already exists",
            detail_pattern=(
                "A workflow template called 'duplicate' already exists"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    @override_permission(Scope, "can_create_workspace", AllowAll)
    def test_post_create_child_workspace(self) -> None:
        """Create a new ``create_child_workspace`` workflow template."""
        static_parameters = {"prefix": "experiment"}
        runtime_parameters: RuntimeParameters = {
            "suffix": RuntimeParameter.ANY,
            "public": RuntimeParameter.ANY,
            "owner_group": RuntimeParameter.ANY,
            "workflow_template_names": RuntimeParameter.ANY,
            "expiration_delay": RuntimeParameter.ANY,
        }

        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="create_child_workspace",
                static_parameters=static_parameters,
                runtime_parameters=runtime_parameters,
            )
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=self.scenario.workspace
        )
        self.assertEqual(template.task_name, "create_child_workspace")
        self.assertEqual(template.static_parameters, static_parameters)
        self.assertEqual(template.runtime_parameters, runtime_parameters)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_different_workspace(self) -> None:
        workspace = self.playground.create_workspace(
            name="test-workspace", public=True
        )
        static_parameters = {"target_distribution": "debian:bookworm"}

        response = self.post(
            workspace=workspace,
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="sbuild",
                static_parameters=static_parameters,
            ),
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=workspace
        )
        self.assertEqual(template.task_name, "sbuild")
        self.assertEqual(template.static_parameters, static_parameters)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_positive_priority_without_permissions(self) -> None:
        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="noop",
                static_parameters={},
                priority=1,
            ),
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="You are not permitted to set positive priorities",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_positive_priority_with_permissions(self) -> None:
        self.playground.add_user_permission(
            self.scenario.user, WorkRequest, "manage_workrequest_priorities"
        )
        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="noop",
                static_parameters={},
                priority=1,
            ),
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=self.scenario.workspace
        )
        self.assertEqual(template.priority, 1)

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_extra_groups(self) -> None:
        group1 = self.playground.create_group("group1")
        group2 = self.playground.create_group("group2")
        group1.users.add(self.scenario.user)
        group2.users.add(self.scenario.user)
        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="noop",
                static_parameters={},
                extra_groups=[group1.pk, str(group2)],
            ),
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=self.scenario.workspace
        )
        self.assertQuerySetEqual(
            template.extra_groups.all(), [group1, group2], ordered=False
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_extra_groups_not_accessible(self) -> None:
        group1 = self.playground.create_group("group1")
        group2 = self.playground.create_group("group2")
        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="noop",
                static_parameters={},
                extra_groups=[group1.pk, str(group2)],
            ),
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern=r"You are not allowed to grant"
            rf" extra group {re.escape(str(group1))}",
            status_code=status.HTTP_403_FORBIDDEN,
        )
        self.assertQuerySetEqual(
            WorkflowTemplate.objects.filter(name="test"), []
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_post_restricted(self) -> None:
        response = self.post(
            data=WorkflowTemplateDataNew(
                name="test",
                task_name="noop",
                static_parameters={},
                restricted=True,
            ),
        )
        self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        template = WorkflowTemplate.objects.get(
            name="test", workspace=self.scenario.workspace
        )
        self.assertTrue(template.restricted)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_noop(self) -> None:
        response = self.patch(data={})
        new_data = self.assertAPIResponseOk(response)
        assert WorkflowTemplateData(**new_data) == self.sample_client_with_links
        self.assertSampleUnchanged()

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_noop_all_set(self) -> None:
        response = self.patch(
            data={
                "name": "sample",
                "task_name": "noop",
                "priority": 0,
                "static_parameters": {},
                "runtime_parameters": RuntimeParameter.ANY,
            },
        )
        new_data = self.assertAPIResponseOk(response)
        assert WorkflowTemplateData(**new_data) == self.sample_client_with_links
        self.assertSampleUnchanged()

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_rename(self) -> None:
        response = self.patch(data={"name": "renamed"})
        new_data = self.assertAPIResponseOk(response)
        changed = self.sample_client.model_dump()
        changed["name"] = "renamed"
        self.sample.refresh_from_db()
        changed["links"] = self.get_links(self.sample)
        assert WorkflowTemplateData(**new_data).model_dump() == changed
        self.assertEqual(self.sample.name, "renamed")
        self.assertEqual(self.sample.task_name, "noop")
        self.assertEqual(self.sample.priority, 0)
        self.assertEqual(self.sample.static_parameters, {})

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_static_parameters(self) -> None:
        static_parameters = {"changed": True}
        response = self.patch(
            workflow_template=self.sample,
            data={"static_parameters": static_parameters},
        )
        self.assertAPIResponseOk(response)
        self.sample.refresh_from_db()
        self.assertEqual(self.sample.static_parameters, static_parameters)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    @override_permission(Scope, "can_create_workspace", AllowAll)
    def test_patch_create_child_workspace(self) -> None:
        template = self.playground.create_workflow_template(
            "test",
            task_name="create_child_workspace",
            static_parameters={"prefix": "experiment"},
        )
        static_parameters = {"prefix": "new-experiment"}
        response = self.patch(
            workflow_template=template,
            data={"static_parameters": static_parameters},
        )
        self.assertAPIResponseOk(response)
        template.refresh_from_db()
        self.assertEqual(template.static_parameters, static_parameters)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_positive_priority_without_permissions(self) -> None:
        response = self.patch(
            workflow_template=self.sample, data={"priority": 1}
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="You are not permitted to set positive priorities",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_positive_priority_with_permissions(self) -> None:
        self.playground.add_user_permission(
            self.scenario.user, WorkRequest, "manage_workrequest_priorities"
        )
        response = self.patch(
            workflow_template=self.sample, data={"priority": 1}
        )
        self.assertAPIResponseOk(response)
        self.sample.refresh_from_db()
        self.assertEqual(self.sample.priority, 1)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_restricted(self) -> None:
        response = self.patch(
            workflow_template=self.sample, data={"restricted": True}
        )
        self.assertAPIResponseOk(response)
        self.sample.refresh_from_db()
        self.assertTrue(self.sample.restricted)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_change_all(self) -> None:
        response = self.patch(
            data={
                "name": "renamed",
                "priority": -3,
                "task_name": "noop",
                "static_parameters": {"changed": True},
                "runtime_parameters": {},
                "restricted": True,
            }
        )
        new_data = self.assertAPIResponseOk(response)
        self.sample.refresh_from_db()
        assert new_data == {
            "id": self.sample.pk,
            "name": "renamed",
            "task_name": "noop",
            "priority": -3,
            "static_parameters": {"changed": True},
            "runtime_parameters": {},
            "links": self.get_links(self.sample),
            "extra_groups": [],
            "restricted": True,
        }
        self.assertEqual(self.sample.name, "renamed")
        self.assertEqual(self.sample.task_name, "noop")
        self.assertEqual(self.sample.priority, -3)
        self.assertEqual(self.sample.static_parameters, {"changed": True})
        self.assertEqual(self.sample.runtime_parameters, {})
        self.assertTrue(self.sample.restricted)

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_cannot_change_task_name(self) -> None:
        response = self.patch(
            workflow_template=self.sample, data={"task_name": "sbuild"}
        )
        self.assertResponseProblem(
            response,
            "The task name of a workflow template cannot be changed",
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertSampleUnchanged()

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_rename_conflict(self) -> None:
        self.playground.create_workflow_template("renamed", task_name="noop")

        response = self.patch(data={"name": "renamed"})
        self.assertResponseProblem(
            response,
            "New name conflicts with an existing workflow template",
            detail_pattern=(
                "A workflow template called 'renamed' already exists"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )
        self.assertSampleUnchanged()

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_rename_no_conflict_in_other_workspaces(self) -> None:
        other = self.playground.create_workspace(name="other")
        self.playground.create_workflow_template(
            "renamed", task_name="noop", workspace=other
        )

        response = self.patch(data={"name": "renamed"})
        new_data = self.assertAPIResponseOk(response)
        changed = self.sample_client.model_copy()
        changed.name = "renamed"
        self.sample.refresh_from_db()
        changed.links = self.get_links(self.sample)
        assert WorkflowTemplateData(**new_data) == changed

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_extra_groups(self) -> None:
        group1 = self.playground.create_group("group1")
        group2 = self.playground.create_group("group2")
        group1.users.add(self.scenario.user)
        group2.users.add(self.scenario.user)
        response = self.patch(
            workflow_template=self.sample,
            data={"extra_groups": [group1.pk, str(group2)]},
        )
        new_data = self.assertAPIResponseOk(response)
        self.sample.refresh_from_db()
        assert new_data == {
            "id": self.sample.pk,
            "name": "sample",
            "task_name": "noop",
            "priority": 0,
            "static_parameters": {},
            "runtime_parameters": "any",
            "links": self.get_links(self.sample),
            "extra_groups": [str(group1), str(group2)],
            "restricted": False,
        }
        self.assertEqual(self.sample.name, "sample")
        self.assertEqual(self.sample.task_name, "noop")
        self.assertEqual(self.sample.priority, 0)
        self.assertEqual(self.sample.static_parameters, {})
        self.assertEqual(self.sample.runtime_parameters, "any")
        self.assertQuerySetEqual(
            self.sample.extra_groups.all(), [group1, group2], ordered=False
        )

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_patch_cannot_set_unaccessible_extra_groups(self) -> None:
        group = self.playground.create_group("group")
        response = self.patch(
            workflow_template=self.sample, data={"extra_groups": [group.pk]}
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern=r"You are not allowed to grant"
            rf" extra group {re.escape(str(group))}",
            status_code=status.HTTP_403_FORBIDDEN,
        )
        self.assertSampleUnchanged()

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_debusine_workflow_template_create(self) -> None:
        debusine = self.get_debusine()
        request = WorkflowTemplateDataNew(
            name="test",
            task_name="noop",
            static_parameters={},
            priority=0,
        )
        result = debusine.workflow_template_create(
            self.scenario.workspace.name, request
        )
        created = WorkflowTemplate.objects.get(
            workspace=self.scenario.workspace, name="test"
        )
        self.assertEqual(result.id, created.id)
        self.assertEqual(result.name, "test")
        self.assertEqual(result.static_parameters, {})
        self.assertEqual(result.priority, 0)

    def test_debusine_workflow_template_get_pk(self) -> None:
        debusine = self.get_debusine()
        result = debusine.workflow_template_get(
            self.scenario.workspace.name, self.sample.pk
        )
        self.assertEqual(result, self.sample_client_with_links)

    def test_debusine_workflow_template_get_name(self) -> None:
        debusine = self.get_debusine()
        result = debusine.workflow_template_get(
            self.scenario.workspace.name, self.sample.name
        )
        self.assertEqual(result, self.sample_client_with_links)

    def test_debusine_workflow_template_iter(self) -> None:
        debusine = self.get_debusine()
        result = debusine.workflow_template_iter(self.scenario.workspace.name)
        self.assertEqual([c.name for c in result], ["sample"])

    @override_permission(WorkflowTemplate, "can_configure", AllowAll)
    def test_debusine_workflow_template_update(self) -> None:
        debusine = self.get_debusine()
        changed = self.sample_client.model_copy()
        changed.name = "renamed"
        changed.priority = -2
        changed.static_parameters = {"changed": True}
        changed.runtime_parameters = RuntimeParameter.ANY
        result = debusine.workflow_template_update(
            self.scenario.workspace.name, changed
        )
        self.sample.refresh_from_db()
        changed_with_links = changed.model_copy()
        changed_with_links.links = self.get_links(self.sample)
        self.assertEqual(result, changed_with_links)
        self.assertEqual(self.sample.name, "renamed")
        self.assertEqual(self.sample.task_name, "noop")
        self.assertEqual(self.sample.priority, -2)
        self.assertEqual(self.sample.static_parameters, {"changed": True})
        self.assertEqual(self.sample.runtime_parameters, RuntimeParameter.ANY)


class WorkflowViewTests(TestCase):
    """Tests for WorkflowView."""

    scenario = scenarios.DefaultContextAPI()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.client = APIClient()
        self.token = self.scenario.user_token

    def post_workflow(
        self, data: dict[str, Any], scope: str | None = None
    ) -> TestResponseType:
        """Post a workflow creation request to api:workflows."""
        headers = {"Token": self.token.key}
        if scope is not None:
            headers["X-Debusine-Scope"] = scope

        return self.client.post(
            reverse("api:workflows"), data=data, headers=headers, format="json"
        )

    def test_authentication_credentials_not_provided(self) -> None:
        """A Token is required to use the endpoint."""
        response = self.client.post(reverse("api:workflows"))
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="Authentication credentials were not provided.",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_create_workflow_success(self) -> None:
        """Create a workflow."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        artifact = self.playground.create_source_artifact()
        for architecture in ("amd64", "arm64"):
            self.playground.create_debian_environment(
                codename="bookworm", architecture=architecture
            )

        WorkflowTemplate.objects.create(
            name="test",
            workspace=default_workspace(),
            task_name="sbuild",
            static_parameters={"architectures": ["amd64", "arm64"]},
            runtime_parameters=RuntimeParameter.ANY,
        )

        response = self.post_workflow(
            {
                "template_name": "test",
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                },
            }
        )

        data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        workflow = WorkRequest.objects.latest("created_at")
        self.assertEqual(data, WorkRequestSerializer(workflow).data)
        self.assertDictContainsAll(
            data,
            {
                "workspace": settings.DEBUSINE_DEFAULT_WORKSPACE,
                "created_by": self.token.user_id,
                "status": WorkRequest.Statuses.PENDING,
                "task_type": TaskTypes.WORKFLOW,
                "task_name": "sbuild",
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                    "architectures": ["amd64", "arm64"],
                },
                "priority_base": 0,
            },
        )

    def test_create_workflow_different_workspace(self) -> None:
        """Create a workflow in a non-default workspace."""
        workspace = self.playground.create_workspace(
            name="test-workspace", public=True
        )
        self.playground.create_group_role(
            workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        artifact = self.playground.create_source_artifact(workspace=workspace)
        for architecture in ("amd64", "arm64"):
            self.playground.create_debian_environment(
                workspace=workspace,
                codename="bookworm",
                architecture=architecture,
            )
        WorkflowTemplate.objects.create(
            name="test",
            workspace=workspace,
            task_name="sbuild",
            static_parameters={"architectures": ["amd64", "arm64"]},
            runtime_parameters=RuntimeParameter.ANY,
        )

        response = self.post_workflow(
            {
                "template_name": "test",
                "workspace": workspace.name,
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                },
            }
        )

        data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        workflow = WorkRequest.objects.latest("created_at")
        self.assertEqual(data, WorkRequestSerializer(workflow).data)
        self.assertEqual(workflow.workspace, workspace)

    def test_create_workflow_without_template_success(self) -> None:
        """Create a workflow without a template."""
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )
        artifact = self.playground.create_source_artifact()
        for architecture in ("amd64", "arm64"):
            self.playground.create_debian_environment(
                codename="bookworm", architecture=architecture
            )

        response = self.post_workflow(
            {
                "task_name": "sbuild",
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                    "architectures": ["amd64", "arm64"],
                },
            }
        )

        data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        workflow = WorkRequest.objects.latest("created_at")
        self.assertEqual(data, WorkRequestSerializer(workflow).data)
        self.assertDictContainsAll(
            data,
            {
                "workspace": settings.DEBUSINE_DEFAULT_WORKSPACE,
                "created_by": self.token.user_id,
                "status": WorkRequest.Statuses.PENDING,
                "task_type": TaskTypes.WORKFLOW,
                "task_name": "sbuild",
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                    "architectures": ["amd64", "arm64"],
                },
                "priority_base": 0,
            },
        )

    @override_permission(Scope, "can_create_workspace", AllowAll)
    def test_create_child_workspace_without_template_success(self) -> None:
        """Create a ``create_child_workspace`` workflow without a template."""
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post_workflow(
            {
                "task_name": "create_child_workspace",
                "task_data": {"suffix": "test"},
            }
        )

        data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        workflow = WorkRequest.objects.latest("created_at")
        self.assertEqual(data, WorkRequestSerializer(workflow).data)
        self.assertDictContainsAll(
            data,
            {
                "workspace": settings.DEBUSINE_DEFAULT_WORKSPACE,
                "created_by": self.token.user_id,
                "status": WorkRequest.Statuses.PENDING,
                "task_type": TaskTypes.WORKFLOW,
                "task_name": "create_child_workspace",
                "task_data": {"suffix": "test"},
                "priority_base": 0,
            },
        )

    def test_create_workflow_nonexistent_template(self) -> None:
        """The view returns HTTP 404 if the workflow template does not exist."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {"template_name": "test", "task_data": {}}
            )

        self.assertResponseProblem(
            response,
            "Workflow template not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_create_workflow_invalid_task_data(self) -> None:
        """Creating a workflow validates task data."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        self.playground.create_workflow_template(name="noop", task_name="noop")

        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {"template_name": "noop", "task_data": {"nonexistent": True}}
            )

        self.assertResponseProblem(
            response,
            "Cannot create workflow",
            detail_pattern=r"Extra inputs are not permitted",
        )

    def test_create_workflow_thorough_input_validation(self) -> None:
        """Creating a workflow does thorough validation of input data."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        self.playground.create_workflow_template(
            name="sbuild", task_name="sbuild"
        )
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )

        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {
                    "template_name": "sbuild",
                    "task_data": {
                        "input": {"source_artifact": artifact.id},
                        "target_distribution": "debian:bookworm",
                        "architectures": ["amd64"],
                    },
                }
            )

        self.assertResponseProblem(
            response,
            "Cannot create workflow",
            detail_pattern=(
                r"^input.source_artifact: unexpected artifact category: "
                r"'debusine:test'. Valid categories: "
                r"\['debian:source-package', 'debian:upload'\]$"
            ),
        )

    def test_create_workflow_template_validation_untruncated(self) -> None:
        """Pydantic validation messages are not truncated."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        self.playground.create_workflow_template(
            name="sbuild",
            task_name="sbuild",
            static_parameters={
                "very_long_unknown_parameter_name_foo_bar_baz_quux_corge": "1"
            },
        )

        response = self.post_workflow(
            {"template_name": "sbuild", "task_data": {}}
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # pydantic validation messages can get truncated if rended by str()
        self.assertNotIn("...", response.json()["detail"])

    def test_create_workflow_template_task_config_error(self) -> None:
        """non-Pydantic validation messages are handled correctly."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        self.playground.create_workflow_template(
            name="sbuild",
            task_name="sbuild",
            static_parameters={},
        )
        self.playground.create_debian_environment(
            codename="bookworm", architecture="amd64"
        )
        artifact = self.playground.create_source_artifact()

        with mock.patch.object(
            WorkRequest,
            "get_task",
            side_effect=TaskConfigError("Unknown failure"),
        ):
            response = self.post_workflow(
                {
                    "template_name": "sbuild",
                    "task_data": {
                        "input": {"source_artifact": artifact.id},
                        "target_distribution": "debian:bookworm",
                        "architectures": ["amd64"],
                    },
                }
            )

        self.assertResponseProblem(
            response,
            "Cannot create workflow",
            detail_pattern=r"Unknown failure",
        )

    def test_create_workflow_honours_scope(self) -> None:
        """Creating a workflow looks up workspace in current scope."""
        artifact = self.playground.create_source_artifact()
        for architecture in ("amd64", "i386"):
            self.playground.create_debian_environment(
                codename="bookworm", architecture=architecture
            )

        scope1 = self.playground.get_or_create_scope("scope1")
        scope2 = self.playground.get_or_create_scope("scope2")
        scope3 = self.playground.get_or_create_scope("scope3")
        workspace1 = self.playground.create_workspace(
            scope=scope1, name="common-name", public=True
        )
        workspace2 = self.playground.create_workspace(
            scope=scope2, name="common-name", public=True
        )
        for workspace, architectures in (
            (workspace1, ["amd64"]),
            (workspace2, ["i386"]),
        ):
            workspace.set_inheritance([self.scenario.workspace])
            self.playground.create_workflow_template(
                name="test",
                task_name="sbuild",
                workspace=workspace,
                static_parameters={"architectures": architectures},
            )
        task_data = {
            "input": {"source_artifact": artifact.id},
            "target_distribution": "debian:bookworm",
        }

        for workspace, architectures in (
            (workspace1, ["amd64"]),
            (workspace2, ["i386"]),
        ):
            self.playground.create_group_role(
                workspace,
                Workspace.Roles.CONTRIBUTOR,
                [self.scenario.user],
            )

            response = self.post_workflow(
                {
                    "template_name": "test",
                    "workspace": "common-name",
                    "task_data": task_data,
                },
                scope=workspace.scope.name,
            )

            data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
            workflow = WorkRequest.objects.latest("created_at")
            with urlconf_scope(workflow.workspace.scope.name):
                self.assertEqual(data, WorkRequestSerializer(workflow).data)
            self.assertEqual(workflow.workspace, workspace)
            self.assertEqual(workflow.task_data["architectures"], architectures)

        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {
                    "template_name": "test",
                    "workspace": "common-name",
                    "task_data": task_data,
                },
                scope=scope3.name,
            )
        self.assertResponseProblem(
            response,
            "Workspace not found",
            detail_pattern="Workspace common-name not found in scope scope3",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_create_workflow_no_default_workspace(self) -> None:
        """POST with no workspace in a scope without a default workspace."""
        scope = self.playground.get_or_create_scope("empty-scope")

        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {"template_name": "test"}, scope=scope.name
            )

        self.assertResponseProblem(
            response,
            "Cannot deserialize workflow creation request",
            validation_errors_pattern=(
                r"'workspace': \['This field is required\.'\]"
            ),
        )

    def test_create_workflow_private_workspace_unauthorized(self) -> None:
        """POST to private workspaces 404s to the unauthorized."""
        private_workspace = self.playground.create_workspace(name="Private")
        self.playground.create_workflow_template(
            name="test", task_name="noop", workspace=private_workspace
        )

        with self.assert_model_count_unchanged(WorkRequest):
            response = self.post_workflow(
                {"template_name": "test", "workspace": private_workspace.name},
                scope=private_workspace.scope.name,
            )

        self.assertResponseProblem(
            response,
            "Workspace not found",
            detail_pattern=(
                f"Workspace {private_workspace.name} not found in scope "
                f"{private_workspace.scope.name}"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_create_workflow_private_workspace_authorized(self) -> None:
        """POST to private workspaces succeeds for the authorized."""
        private_workspace = self.playground.create_workspace(name="Private")
        self.playground.create_group_role(
            private_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.playground.create_workflow_template(
            name="test", task_name="noop", workspace=private_workspace
        )

        response = self.post_workflow(
            {"template_name": "test", "workspace": private_workspace.name},
            scope=private_workspace.scope.name,
        )

        data = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)
        workflow = WorkRequest.objects.latest("created_at")
        self.assertEqual(data, WorkRequestSerializer(workflow).data)
        self.assertEqual(workflow.workspace, private_workspace)

    def test_create_workflow_without_template_not_owner(self) -> None:
        """Only workspace owners can create workflows without a template."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            [self.scenario.user],
        )
        artifact = self.playground.create_source_artifact()

        response = self.post_workflow(
            {
                "task_name": "sbuild",
                "task_data": {
                    "input": {"source_artifact": artifact.id},
                    "target_distribution": "debian:bookworm",
                    "architectures": ["amd64", "arm64"],
                },
            }
        )

        self.assertResponseProblem(
            response,
            f"playground cannot configure workspace {self.scenario.workspace}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    @override_permission(Scope, "can_create_workspace", DenyAll)
    def test_create_child_workspace_without_template_not_scope_owner(
        self,
    ) -> None:
        """Only scope owners may create ``create_child_workspace`` templates."""
        self.playground.create_group_role(
            self.scenario.workspace, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post_workflow(
            {
                "task_name": "create_child_workspace",
                "task_data": {"suffix": "test"},
            }
        )

        self.assertResponseProblem(
            response,
            f"playground cannot create workspaces in {self.scenario.scope}",
            status_code=status.HTTP_403_FORBIDDEN,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for SystemBootstrap class."""

import copy
import hashlib
import io
import textwrap
from pathlib import Path
from textwrap import dedent
from typing import Any, cast, override
from unittest import mock
from unittest.mock import call

import requests
import responses
from debian.deb822 import Deb822

import debusine.worker.tags as wtags
from debusine.client.exceptions import ContentValidationError
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    BaseDynamicTaskData,
    SystemBootstrapData,
    SystemBootstrapRepository,
    SystemBootstrapRepositoryCheckSignatureWith,
    SystemBootstrapRepositoryKeyring,
    SystemBootstrapRepositoryType,
)
from debusine.tasks.systembootstrap import SystemBootstrap
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class SystemBootstrapImpl(SystemBootstrap[SystemBootstrapData]):
    """Implementation of SystemBootstrap ontology."""

    @override
    def _cmdline(self) -> list[str]:
        return []  # pragma: no cover


class SystemBootstrapTests(
    ExternalTaskHelperMixin[SystemBootstrapImpl], TestCase
):
    """Tests SystemBootstrap class."""

    SAMPLE_TASK_DATA = {
        "bootstrap_options": {
            "architecture": "amd64",
            "variant": "buildd",
            "extra_packages": ["hello", "python3"],
        },
        "bootstrap_repositories": [
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "system",
            },
            {
                "types": ["deb-src"],
                "mirror": "https://example.com",
                "suite": "bullseye",
                "components": ["main"],
                "check_signature_with": "no-check",
            },
        ],
    }

    def test_configure_task(self) -> None:
        """Ensure self.SAMPLE_TASK_DATA is valid."""
        self.configure_task(self.SAMPLE_TASK_DATA)

        # Test default options
        sample_repositories = cast(
            list[dict[str, Any]],
            self.SAMPLE_TASK_DATA["bootstrap_repositories"],
        )
        actual_repositories = self.task.data.bootstrap_repositories

        # First sample repository does not have "types" nor
        # "signed_signature_with"
        self.assertNotIn("types", sample_repositories[0])
        self.assertNotIn("signed_signature_with", sample_repositories[0])
        # They are assigned to the default values
        self.assertEqual(actual_repositories[0].types, ["deb"])
        self.assertEqual(actual_repositories[0].check_signature_with, "system")

        #  Second sample repository has "types" / "signed_signature_with" as
        # per SAMPLE_TASK_DATA
        self.assertEqual(sample_repositories[1]["types"], ["deb-src"])
        self.assertEqual(
            actual_repositories[1].check_signature_with, "no-check"
        )

    def test_configure_task_raise_value_error_missing_keyring(self) -> None:
        """configure_task() raise ValueError: missing keyring parameter."""
        task_data = copy.deepcopy(self.SAMPLE_TASK_DATA)
        bootstrap_repo = cast(
            list[dict[str, Any]],
            copy.deepcopy(self.SAMPLE_TASK_DATA["bootstrap_repositories"]),
        )[0]
        bootstrap_repo["check_signature_with"] = "external"
        task_data["bootstrap_repositories"] = [bootstrap_repo]

        msg = (
            r"repository requires 'keyring':"
            r" 'check_signature_with' is set to 'external'"
        )

        with self.assertRaisesRegex(TaskConfigError, msg):
            self.configure_task(task_data)

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:systembootstrapimpl:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            require_system={wtags.BUILD_ARCH_PREFIX + "amd64"}
        )

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BaseDynamicTaskData(
                subject="bookworm",
                parameter_summary="bookworm (amd64-buildd)",
                runtime_context="amd64:buildd:hello,python3",
                configuration_context="amd64",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "bookworm")

    def test_get_configuration_context(self) -> None:
        self.assertEqual(self.get_configuration_context(), "amd64")

    @staticmethod
    def create_and_validate_repo(
        check_signature_with: SystemBootstrapRepositoryCheckSignatureWith,
        *,
        keyring_url: str | None = None,
        keyring_sha256sum: str = "",
    ) -> SystemBootstrapRepository:
        """Create a basic repository, validates and return it."""
        keyring: SystemBootstrapRepositoryKeyring | None = None
        if keyring_url is not None:
            keyring = SystemBootstrapRepositoryKeyring(
                url=keyring_url, sha256sum=keyring_sha256sum
            )
        repository = SystemBootstrapRepository(
            types=[SystemBootstrapRepositoryType.DEB],
            mirror="https://example.com",
            suite="bullseye",
            components=["main", "contrib"],
            check_signature_with=check_signature_with,
            keyring=keyring,
        )

        return repository

    def test_deb822_source(self) -> None:
        """Generate a correct Deb822 sources file."""
        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.SYSTEM
        )

        actual = self.task._deb822_source(
            repository,
            keyring_directory=self.create_temporary_directory(),
            use_signed_by=True,
        )

        expected = {
            "Types": "deb",
            "URIs": "https://example.com",
            "Suites": "bullseye",
            "Components": "main contrib",
        }

        self.assertEqual(actual, expected)

    def test_deb822_source_no_components(self) -> None:
        """Use components returned by _list_components_for_suite."""
        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.NO_CHECK
        )
        repository.components = None

        components = ["main", "contrib"]
        with mock.patch(
            "debusine.tasks.systembootstrap.SystemBootstrap."
            "_list_components_for_suite",
            return_value=components,
            autospec=True,
        ) as mocked:
            actual = self.task._deb822_source(
                repository,
                keyring_directory=self.create_temporary_directory(),
                use_signed_by=False,
            )

        self.assertEqual(actual["components"], " ".join(components))

        mocked.assert_called_with(
            repository.mirror, repository.suite, timeout=DEFAULT_REQUEST_TIMEOUT
        )

    def test_deb822_source_signature_no_check(self) -> None:
        """Generate Deb822 with "trusted"."""
        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.NO_CHECK
        )

        actual = self.task._deb822_source(
            repository,
            keyring_directory=self.create_temporary_directory(),
            use_signed_by=True,
        )

        self.assertEqual(actual["Trusted"], "yes")

    @responses.activate
    def test_deb822_source_file_download_keyring_binary(self) -> None:
        """Generate Deb822 and downloads a binary keyring file."""
        keyring_contents = b"\x99\x02"
        keyring_url = "https://example.com/keyring_file.gpg"
        keyring_directory = self.create_temporary_directory()

        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL,
            keyring_url=keyring_url,
        )

        self.assertIsNotNone(repository.keyring)
        assert repository.keyring
        responses.add(
            responses.GET, str(repository.keyring.url), body=keyring_contents
        )

        actual = self.task._deb822_source(
            repository, keyring_directory=keyring_directory, use_signed_by=True
        )

        signed_by_keyring = Path(actual["Signed-By"])
        self.assertTrue(signed_by_keyring.name.endswith(".gpg"))
        self.assertEqual(signed_by_keyring.read_bytes(), keyring_contents)

        # It was created in the correct directory
        self.assertTrue(signed_by_keyring.is_relative_to(keyring_directory))

    @responses.activate
    def test_deb822_source_file_download_keyring_ascii(self) -> None:
        """Generate Deb822 and downloads an ASCII-armored keyring file."""
        keyring_contents = b"-----BEGIN PGP PUBLIC KEY BLOCK-----\n"
        keyring_url = "https://example.com/keyring_file.asc"
        keyring_directory = self.create_temporary_directory()

        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL,
            keyring_url=keyring_url,
        )

        self.assertIsNotNone(repository.keyring)
        assert repository.keyring
        responses.add(
            responses.GET, str(repository.keyring.url), body=keyring_contents
        )

        actual = self.task._deb822_source(
            repository, keyring_directory=keyring_directory, use_signed_by=True
        )

        signed_by_keyring = Path(actual["Signed-By"])
        self.assertTrue(signed_by_keyring.name.endswith(".asc"))
        self.assertEqual(signed_by_keyring.read_bytes(), keyring_contents)

        # It was created in the correct directory
        self.assertTrue(signed_by_keyring.is_relative_to(keyring_directory))

    @responses.activate
    def test_deb822_source_keyring_hash_mismatch(self) -> None:
        """Method raise ValueError: keyring sha256 mismatch."""
        keyring_contents = b"some-keyring"
        keyring_url = "https://example.com/keyring_file.txt"
        keyring_directory = self.create_temporary_directory()

        expected_sha256 = "will be a mismatch"

        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL,
            keyring_url=keyring_url,
            keyring_sha256sum=expected_sha256,
        )

        self.assertIsNotNone(repository.keyring)
        assert repository.keyring

        responses.add(
            responses.GET, str(repository.keyring.url), body=keyring_contents
        )

        actual_sha256 = hashlib.sha256(keyring_contents).hexdigest()

        expected = (
            f"sha256 mismatch for keyring repository {repository.mirror}. "
            f"Actual: {actual_sha256} expected: {expected_sha256}"
        )

        with self.assertRaisesRegex(ContentValidationError, expected):
            self.task._deb822_source(
                repository,
                keyring_directory=keyring_directory,
                use_signed_by=True,
            )

    @responses.activate
    def test_deb822_source_keyring_hash_correct(self) -> None:
        """Keyring sha256 is correct."""
        keyring_contents = b"some-keyring"
        keyring_url = "https://example.com/keyring_file.txt"
        keyring_directory = self.create_temporary_directory()

        actual_sha256 = hashlib.sha256(keyring_contents).hexdigest()

        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL,
            keyring_url=keyring_url,
            keyring_sha256sum=actual_sha256,
        )

        self.assertIsNotNone(repository.keyring)
        assert repository.keyring

        responses.add(
            responses.GET, str(repository.keyring.url), body=keyring_contents
        )

        self.task._deb822_source(
            repository, keyring_directory=keyring_directory, use_signed_by=True
        )

    def test_deb822_source_keyring_file(self) -> None:
        """The keyring may be read from a file in /usr/share/keyrings/."""
        keyring_contents = b"some-keyring"
        keyring_url = "file:///usr/share/keyrings/debian-archive-keyring.gpg"
        keyring_directory = self.create_temporary_directory()

        repository = self.create_and_validate_repo(
            SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL,
            keyring_url=keyring_url,
        )

        self.assertIsNotNone(repository.keyring)
        assert repository.keyring

        response = requests.Response()
        response.raw = io.BytesIO(keyring_contents)
        response.status_code = requests.codes.ok
        response.url = keyring_url
        with mock.patch(
            "requests_file.FileAdapter.send", return_value=response
        ):
            actual = self.task._deb822_source(
                repository,
                keyring_directory=keyring_directory,
                use_signed_by=True,
            )

        signed_by_keyring = Path(actual["Signed-By"])
        self.assertEqual(signed_by_keyring.read_bytes(), keyring_contents)

        # It was created in the correct directory
        self.assertTrue(signed_by_keyring.is_relative_to(keyring_directory))

    @responses.activate
    def test_generate_deb822_sources_file(self) -> None:
        """Assert _generate_deb822_sources_file() return the expected files."""
        keyring_url = "https://example.com/keyring_1.gpg"
        responses.add(responses.GET, keyring_url)

        repository = SystemBootstrapRepository(
            types=[SystemBootstrapRepositoryType.DEB],
            mirror="https://deb.debian.org/deb",
            suite="bookworm",
            components=["main", "contrib"],
            check_signature_with=(
                SystemBootstrapRepositoryCheckSignatureWith.EXTERNAL
            ),
            keyring=SystemBootstrapRepositoryKeyring(url=keyring_url),
        )

        for use_signed_by in [True, False]:
            with self.subTest(use_signed_by=use_signed_by):
                directory = self.create_temporary_directory()

                sources = self.task._generate_deb822_sources(
                    [repository],
                    keyrings_dir=directory,
                    use_signed_by=use_signed_by,
                )

                self.assertEqual(sources[0]["URIs"], repository.mirror)
                self.assertEqual(len(sources), 1)

                if use_signed_by:
                    self.assertEqual(
                        sources[0]["Signed-By"], str(next(directory.iterdir()))
                    )
                else:
                    self.assertNotIn("Signed-By", sources[0])

    @staticmethod
    def add_responses_release_file(
        components: list[str],
    ) -> tuple[str, str, str]:
        """
        Add a mock response for a Release file with Components: line.

        :returns: mirror_url, suite, release_url.
        """
        mirror_url = "http://deb.debian.org/debian"
        suite = "bookworm"

        release_url = f"{mirror_url}/dists/{suite}/Release"

        responses.add(
            responses.GET,
            release_url,
            body=textwrap.dedent(
                f"""\
                Origin: Debian
                Label: Debian
                Suite: stable
                Version: 12.4
                Codename: bookworm
                Date: Sun, 10 Dec 2023 17:43:24 UTC
                Acquire-By-Hash: yes
                No-Support-for-Architecture-all: Packages
                Components: {' '.join(components)}
                Description: Debian 12.4 Released 10 December 2023
                MD5Sum:
                 0ed6d4c8891eb86358b94bb35d9e4da4  1484322 contrib/Contents-all
                 d0a0325a97c42fd5f66a8c3e29bcea64    981 contrib/Contents-all.gz
                 58f32d515c66daafcdac2595fc984814   84179 contrib/Contents-amd64
                """
            ),
            status=200,
        )

        return mirror_url, suite, release_url

    @responses.activate
    def test_list_components_for_suite(self) -> None:
        """_list_components_for_suite get Release URL and return components."""
        expected_components = [
            "main",
            "updates/main",
            "updates/main/something",
            "contrib",
            "non-free-firmware",
            "non-free",
        ]

        mirror_url, suite, _ = self.add_responses_release_file(
            expected_components
        )

        actual_components = SystemBootstrap._list_components_for_suite(
            mirror_url, suite
        )
        self.assertEqual(actual_components, expected_components)

    @responses.activate
    def test_list_components_for_suite_raise_value_error(self) -> None:
        """
        _list_components_for_suite get Release URL and raise exception.

        The line containing "Components:" was not found.
        """
        mirror_url = "http://deb.debian.org/debian"

        url = mirror_url + "/dists/bookworm/Release"

        responses.add(
            responses.GET,
            url,
            body="Some text not containing line starts with Component: ",
            status=200,
        )

        msg = f"^Cannot find components in {url}"
        with self.assertRaisesRegex(ValueError, msg):
            SystemBootstrap._list_components_for_suite(mirror_url, "bookworm")

    @responses.activate
    def test_list_components_raise_value_error_invalid_character_dot(
        self,
    ) -> None:
        """
        _list_components_for_suite raise ValueError: invalid component name.

        Components cannot have a ".".
        """
        component = "non.free"

        mirror_url, suite, release_url = self.add_responses_release_file(
            [component]
        )

        msg = (
            rf'^Invalid component name from {release_url}: "{component}" '
            r'must start with \[A-Za-z\] and have only \[-/A-Za-z\] characters'
        )
        with self.assertRaisesRegex(ValueError, msg):
            SystemBootstrap._list_components_for_suite(mirror_url, suite)

    @responses.activate
    def test_list_components_raise_value_error_invalid_start_with_slash(
        self,
    ) -> None:
        """
        _list_components_for_suite raise ValueError: invalid component name.

        Components cannot start with "/".
        """
        component = "/main"

        mirror_url, suite, release_url = self.add_responses_release_file(
            [component]
        )

        msg = (
            rf'^Invalid component name from {release_url}: "{component}" '
            r'must start with \[A-Za-z\] and have only \[-/A-Za-z\] characters'
        )
        with self.assertRaisesRegex(ValueError, msg):
            SystemBootstrap._list_components_for_suite(mirror_url, suite)

    def test_get_source_components_reads_first_source(self) -> None:
        sources = self.create_temporary_file(
            contents=dedent(
                """\
                Types: deb
                URIs: http://deb.debian.org/debian
                Suites: unstable
                Components: main contrib

                Types: deb
                URIs: http://deb.debian.org/debian
                Suites: unstable
                Components: non-free
                """
            ).encode("ascii")
        )
        self.assertEqual(
            SystemBootstrap._get_source_components(sources), ["main", "contrib"]
        )

    def test_get_source_components_returns_empty_list(self) -> None:
        sources = self.create_temporary_file(contents=b"")
        self.assertEqual(SystemBootstrap._get_source_components(sources), [])

    @responses.activate
    def test_configure_for_execution(self) -> None:
        """Test configure_for_execution(): call _generate_deb822_sources."""
        download_dir = self.create_temporary_directory()

        # Assume that, in production code, the directory was created
        # with restrictive permissions
        download_dir.chmod(0o700)

        url_0 = "https://example.com/keyring1.asc"
        url_1 = "https://example.com/keyring2.asc"

        repositories: list[dict[str, Any]] = [
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "system",
            },
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "no-check",
            },
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "external",
                "keyring": {"url": url_0, "install": True},
            },
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bullseye",
                "components": ["main"],
                "check_signature_with": "external",
                "keyring": {"url": url_1, "install": False},
            },
        ]

        keyring_contents_0 = (
            b"-----BEGIN PGP PUBLIC KEY BLOCK-----\nASCII-armored keyring"
        )
        keyring_contents_1 = b"Binary keyring"

        responses.add(
            responses.GET,
            url_0,
            body=keyring_contents_0,
        )
        responses.add(
            responses.GET,
            url_1,
            body=keyring_contents_1,
        )

        self.configure_task(override={"bootstrap_repositories": repositories})

        with mock.patch.object(
            self.task, "append_to_log_file", autospec=True
        ) as append_to_log_file:
            self.assertTrue(self.task.configure_for_execution(download_dir))

        # Files were generated
        sources_files = list(download_dir.glob("*.sources"))
        self.assertEqual(len(sources_files), 2)

        assert self.task._chroot_sources_file
        # Source file Signed-By changed to the correct chroot path
        with self.task._chroot_sources_file.open() as f:
            for actual, task_repository in zip(
                Deb822.iter_paragraphs(f), repositories
            ):
                if task_repository["check_signature_with"] in (
                    "system",
                    "no-check",
                ):
                    # For a "system" or "no-check": the chroot sources list
                    # does not have any "Signed-By"
                    self.assertNotIn("Signed-By", actual)
                else:
                    # It must be "external"
                    self.assertEqual(
                        task_repository["check_signature_with"], "external"
                    )

                    if task_repository["keyring"]["install"]:
                        # The path changed to one that will make sense in chroot
                        self.assertRegex(
                            actual["Signed-By"],
                            r"^/etc/apt/keyrings-debusine/.*\.asc",
                        )
                    else:
                        # No Signed-By in the chroot: the keyring is not
                        # uploaded (it is used for the initial setup only)
                        self.assertNotIn("Signed-By", actual)

        # check files were added in the log artifact
        expected_calls = [
            call(source_file.name, source_file.read_text().splitlines())
            for source_file in [
                download_dir / "host.sources",
                download_dir / "chroot.sources",
            ]
        ]
        append_to_log_file.assert_has_calls(expected_calls)

        assert self.task._host_sources_file
        host_repositories = list(
            Deb822.iter_paragraphs(self.task._host_sources_file.read_text())
        )

        # Both repositories with check_signature_external have Signed-By
        # in the host.sources
        self.assertEqual(
            Path(host_repositories[2]["Signed-By"]).read_bytes(),
            keyring_contents_0,
        )
        self.assertEqual(
            Path(host_repositories[3]["Signed-By"]).read_bytes(),
            keyring_contents_1,
        )

        # And the files were added in self.task._keyrings
        self.assertIn(
            Path(host_repositories[2]["Signed-By"]), self.task._keyrings
        )
        self.assertIn(
            Path(host_repositories[3]["Signed-By"]), self.task._keyrings
        )

        # With appropriate extensions
        self.assertEqual(Path(host_repositories[2]["Signed-By"]).suffix, ".asc")
        self.assertEqual(Path(host_repositories[3]["Signed-By"]).suffix, ".gpg")

        # Two keyrings added
        self.assertEqual(len(self.task._keyrings), 2)

        # Verify that self.task._upload_keyrings contains only one, correct
        # suffix, correct content
        upload_keyrings = self.task._upload_keyrings
        self.assertEqual(len(upload_keyrings), 1)
        self.assertEqual(upload_keyrings[0].suffix, ".asc")
        self.assertEqual(upload_keyrings[0].read_bytes(), keyring_contents_0)

        # The chroot sources file have one Signed-By and the other on
        # no Signed-By
        host_repositories = list(
            Deb822.iter_paragraphs(self.task._chroot_sources_file.read_text())
        )
        self.assertRegex(
            host_repositories[2]["Signed-By"],
            r"^/etc/apt/keyrings-debusine/keyring-repo-.*\.asc$",
        )
        self.assertNotIn("Signed-By", host_repositories[3])

        # Download directory can be read by anyone (needed for mmdebstrap to
        # use the keys from the subuid, in mode unshare as it is used)
        self.assertEqual(download_dir.stat().st_mode & 0o755, 0o755)

        keyrings_dir = download_dir / "keyrings"
        self.assertEqual(keyrings_dir.stat().st_mode & 0o755, 0o755)

    @responses.activate
    def test_configure_for_execution_no_signed_by_in_repositories(self) -> None:
        """Test configure_for_execution(): use_signed_by is False."""
        download_dir = self.create_temporary_directory()

        bootstrap_options = {"architecture": "amd64", "use_signed_by": False}

        url = "https://example.com/keyring1.asc"

        repositories = [
            {
                "mirror": "https://deb.debian.org/deb",
                "suite": "bookworm",
                "components": ["main", "contrib"],
                "check_signature_with": "external",
                "keyring": {"url": url},
            },
        ]

        responses.add(
            responses.GET,
            url,
        )

        self.configure_task(
            override={
                "bootstrap_repositories": repositories,
                "bootstrap_options": bootstrap_options,
            }
        )
        self.task.configure_for_execution(download_dir)

        assert self.task._chroot_sources_file
        assert self.task._host_sources_file
        # Source file Signed-By changed to the correct chroot path
        with (
            self.task._chroot_sources_file.open() as chroot_sources,
            self.task._host_sources_file.open() as host_sources,
        ):
            count = 0
            for repository in [chroot_sources, host_sources]:
                self.assertNotIn("Signed-By", repository)
                count += 1

            # Expect two: one from chroot, one from host sources
            self.assertEqual(count, 2)

        self.assertEqual(len(self.task._keyrings), 1)
        self.assertEqual(len(self.task._upload_keyrings), 0)

    def test_configure_for_execution_customization_script(self) -> None:
        """Test configure_for_execution() write customization_script."""
        download_dir = self.create_temporary_directory()

        script = "#!/bin/sh\n\necho 'something'\n"
        self.configure_task(
            override={
                "customization_script": script,
            }
        )

        self.task.configure_for_execution(download_dir)

        assert self.task._customization_script
        self.assertEqual(
            self.task._customization_script.name, "customization_script"
        )
        self.assertEqual(self.task._customization_script.read_text(), script)

    def test_get_value_os_release(self) -> None:
        """get_value_os_release() get the correct value."""
        directory = self.create_temporary_directory()
        os_release_data = self.write_os_release(
            directory / self.task._OS_RELEASE_FILE
        )

        actual = SystemBootstrap._get_value_os_release(
            directory / self.task._OS_RELEASE_FILE, "PRETTY_NAME"
        )
        self.assertEqual(actual, os_release_data["PRETTY_NAME"])

    def test_get_value_os_release_key_error(self) -> None:
        """get_value_os_release() raise KeyError."""
        os_release = self.create_temporary_file()

        key = "PRETTY_NAME"
        with self.assertRaisesRegex(KeyError, key):
            SystemBootstrap._get_value_os_release(os_release, key)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to create work requests.

This has fewer restrictions on task types or event reactions than the API
endpoint, and doesn't require a token so it's easier to use in simple
automation.
"""

import argparse
import sys
from typing import Any, NoReturn, override

import yaml
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management import CommandError, CommandParser
from django.db import transaction

from debusine.artifacts.models import TaskTypes
from debusine.db.models import SYSTEM_USER_NAME, User, WorkRequest
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import get_workspace
from debusine.tasks.models import EventReactions


class Command(DebusineBaseCommand):
    """Command to create a work request."""

    help = "Create a new work request"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "task_type",
            choices=[
                TaskTypes.WORKER.lower(),
                TaskTypes.SERVER.lower(),
                TaskTypes.SIGNING.lower(),
            ],
            help="Task type",
        )
        parser.add_argument("task_name", help="Task name")
        parser.add_argument(
            "--created-by",
            help="Name of the user creating the work request",
            default=SYSTEM_USER_NAME,
        )
        parser.add_argument(
            "--workspace",
            metavar="scope/name",
            help="Workspace",
            default=settings.DEBUSINE_DEFAULT_WORKSPACE,
        )
        parser.add_argument(
            "--data",
            type=argparse.FileType("r"),
            help=(
                "File path (or - for stdin) to read the data for the work "
                "request. YAML format. Defaults to stdin."
            ),
            default="-",
        )
        parser.add_argument(
            "--event-reactions",
            type=argparse.FileType("r"),
            help="File path in YAML format requesting notifications.",
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if options["data"] != sys.stdin:
            options["data"].close()
        if options["event_reactions"] not in {None, sys.stdin}:
            options["event_reactions"].close()

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Create the work request."""
        task_type = TaskTypes[options["task_type"].upper()]
        assert task_type in (
            TaskTypes.WORKER,
            TaskTypes.SERVER,
            TaskTypes.SIGNING,
        )

        task_name = options["task_name"]
        created_by_name = options["created_by"]
        data = self.parse_yaml_data(options["data"].read()) or {}
        if options["event_reactions"] is not None:
            if task_type in (TaskTypes.SERVER, TaskTypes.SIGNING):
                raise CommandError(
                    "Event reactions are not supported"
                    f" for {task_type.name} tasks"
                )
            event_reactions = EventReactions(
                **(
                    self.parse_yaml_data(options["event_reactions"].read())
                    or {}
                )
            )
        else:
            event_reactions = EventReactions()

        try:
            created_by = get_user_model().objects.get(username=created_by_name)
        except User.DoesNotExist:
            raise CommandError(
                f'User "{created_by_name}" not found', returncode=3
            )

        workspace = get_workspace(options["workspace"])

        with transaction.atomic():
            try:
                match task_type:
                    case TaskTypes.WORKER:
                        work_request = WorkRequest.objects.create_worker(
                            workspace=workspace,
                            created_by=created_by,
                            task_name=task_name,
                            task_data=data,
                            event_reactions=event_reactions,
                        )
                    case TaskTypes.SERVER:
                        work_request = WorkRequest.objects.create_server(
                            workspace=workspace,
                            created_by=created_by,
                            task_name=task_name,
                            task_data=data,
                        )
                    case TaskTypes.SIGNING:
                        work_request = WorkRequest.objects.create_signing(
                            workspace=workspace,
                            created_by=created_by,
                            task_name=task_name,
                            task_data=data,
                        )
                    case _:  # pragma: no cover
                        # This should already be caught by the assertion at the
                        # top of the method, but mypy doesn't seem to
                        # understand it
                        assert False  # pragma: no cover
                work_request.mark_pending()
            except Exception as e:
                raise CommandError(
                    f"Failed to create work request: {e}", returncode=3
                )
            else:
                print("---")
                print(yaml.safe_dump({"work_request_id": work_request.id}))

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: lookups."""

import logging
from collections.abc import Sequence

from rest_framework import status
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.db.context import context
from debusine.db.models.lookups import LookupResult
from debusine.server.collections.lookup import lookup_multiple, lookup_single
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    LookupMultipleSerializer,
    LookupResponseSerializer,
    LookupSingleSerializer,
)
from debusine.server.views.base import BaseAPIView
from debusine.server.views.rest import IsUserTokenAuthenticated
from debusine.tasks.models import LookupMultiple

logger = logging.getLogger(__name__)


class LookupSingleView(BaseAPIView):
    """View used to look up a single collection item."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]

    def post(self, request: Request) -> Response:
        """
        Resolve a lookup for a single collection item.

        The request data must be a :ref:`lookup-single`.
        """
        lookup_deserialized = LookupSingleSerializer(data=request.data)
        lookup_deserialized.is_valid(raise_exception=True)

        workspace = lookup_deserialized.validated_data["workspace"]
        self.enforce(workspace.can_display)

        if workflow_root := lookup_deserialized.validated_data.get(
            "workflow_root"
        ):
            self.enforce(workflow_root.can_display)

        try:
            result = lookup_single(
                lookup_deserialized.validated_data["lookup"],
                default_category=lookup_deserialized.validated_data.get(
                    "default_category"
                ),
                workspace=workspace,
                pc=context.pc,
                workflow_root=workflow_root,
                expect_type=lookup_deserialized.validated_data["expect_type"],
            )
        except KeyError as e:
            raise DebusineAPIException(
                title="No matches",
                detail=str(e),
                status_code=status.HTTP_404_NOT_FOUND,
            )
        except LookupError as e:
            raise DebusineAPIException(
                title="Lookup error",
                detail=str(e),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        return Response(
            LookupResponseSerializer(result).data, status=status.HTTP_200_OK
        )


class LookupMultipleView(BaseAPIView):
    """View used to look up multiple collection items."""

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]

    def post(self, request: Request) -> Response:
        """
        Resolve a lookup for a single collection item.

        The request data must be a :ref:`lookup-multiple`.
        """
        lookup_deserialized = LookupMultipleSerializer(data=request.data)
        lookup_deserialized.is_valid(raise_exception=True)

        workspace = lookup_deserialized.validated_data["workspace"]
        self.enforce(workspace.can_display)

        if workflow_root := lookup_deserialized.validated_data.get(
            "workflow_root"
        ):
            self.enforce(workflow_root.can_display)

        try:
            lookup = LookupMultiple.model_validate(
                lookup_deserialized.validated_data["lookup"]
            )
        except ValueError as e:
            raise DebusineAPIException(
                title="Cannot deserialize lookup",
                validation_errors={"lookup": str(e)},
            )

        try:
            result: Sequence[LookupResult] = lookup_multiple(
                lookup,
                default_category=lookup_deserialized.validated_data.get(
                    "default_category"
                ),
                workspace=workspace,
                pc=context.pc,
                workflow_root=workflow_root,
                expect_type=lookup_deserialized.validated_data["expect_type"],
            )
        except KeyError as e:
            raise DebusineAPIException(
                title="One of the lookups returned no matches",
                detail=str(e),
                status_code=status.HTTP_404_NOT_FOUND,
            )
        except LookupError as e:
            raise DebusineAPIException(
                title="Lookup error",
                detail=str(e),
                status_code=status.HTTP_400_BAD_REQUEST,
            )

        # The type stubs don't mandle many=True:
        # https://github.com/typeddjango/djangorestframework-stubs/issues/260
        return Response(
            LookupResponseSerializer(
                result,  # type: ignore[arg-type]
                many=True,
            ).data,
            status=status.HTTP_200_OK,
        )



# Generated by Django 4.2.23 on 2025-12-10 10:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0034_workflow_template_configure_runtime_parameters'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='status',
            field=models.CharField(
                choices=[
                    ('pending', 'Pending'),
                    ('running', 'Running'),
                    ('completed', 'Completed'),
                    ('aborted', 'Aborted'),
                    ('blocked', 'Blocked'),
                ],
                default='blocked',
                editable=False,
                max_length=9,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_work_requests."""

import datetime as dt
from typing import ClassVar, override

from django.utils import timezone

from debusine.db.models import WorkRequest
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ListWorkRequestsCommandTests(TabularOutputTests, TestCase):
    """Test for list_work_requests management command.."""

    work_request: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.work_request = cls.playground.create_worker_task(task_name="noop")

    def test_list_work_requests_not_assigned(self) -> None:
        """Test a non-assigned work request output."""
        with self.assertPrintsTable() as output:
            call_command("list_work_requests")
        self.assertEqual(output.col(0), [str(self.work_request.id)])
        self.assertEqual(
            output.col(2), [self.work_request.created_at.isoformat()]
        )
        self.assertEqual(output.col(5), [self.work_request.status])

    def test_list_work_requests_assigned_finished(self) -> None:
        """Test an assigned work request output."""
        worker = self.playground.create_worker(fqdn="neptune")
        self.work_request.worker = worker
        self.work_request.created_at = timezone.now()
        one_sec = dt.timedelta(seconds=1)
        self.work_request.started_at = self.work_request.created_at + one_sec
        self.work_request.completed_at = self.work_request.started_at + one_sec
        self.work_request.result = WorkRequest.Results.SUCCESS
        self.work_request.save()

        with self.assertPrintsTable() as output:
            call_command("list_work_requests")

        self.assertEqual(output.col(1), ["neptune"])
        self.assertEqual(
            output.col(2), [self.work_request.created_at.isoformat()]
        )
        self.assertEqual(
            output.col(3), [self.work_request.started_at.isoformat()]
        )
        self.assertEqual(
            output.col(4), [self.work_request.completed_at.isoformat()]
        )
        self.assertEqual(output.col(6), [self.work_request.result])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for cloud-init script generation."""

from base64 import b64decode
from textwrap import dedent

import yaml
from django.test import SimpleTestCase

from debusine.assets.models import (
    BasicAPTAuthenticationData,
    BasicAPTCredentials,
)
from debusine.db.models import Token
from debusine.server.worker_pools.cloud_init import (
    AptConfig,
    CloudInitConfig,
    FilesystemSetup,
    WriteFile,
    apt_source,
    worker_bootstrap_cloud_init,
)
from debusine.server.worker_pools.models import (
    DebianRelease,
    DebusineInstallSource,
)
from debusine.tasks.models import ExtraExternalRepository
from debusine.test.django import TestCase


class APTSourceTests(SimpleTestCase):
    """Tests for apt_source."""

    def test_daily_builds(self) -> None:
        source = apt_source(DebusineInstallSource.DAILY_BUILDS)
        self.assertEqual(source.components, ["main"])
        self.assertEqual(source.suite, "trixie-devel")
        self.assertEqual(
            str(source.url),
            "https://deb.debusine.freexian.com/freexian/debusine",
        )
        assert source.signing_key is not None
        self.assertTrue(
            source.signing_key.startswith(
                "-----BEGIN PGP PUBLIC KEY BLOCK-----"
            )
        )

    def test_daily_builds_specified_release(self) -> None:
        source = apt_source(
            DebusineInstallSource.DAILY_BUILDS, release=DebianRelease.FORKY
        )
        self.assertEqual(source.suite, "forky-devel")

    def test_enterprise_staging(self) -> None:
        source = apt_source(DebusineInstallSource.ENTERPRISE_STAGING)
        self.assertEqual(source.components, ["main"])
        self.assertEqual(source.suite, "trixie-staging")
        self.assertEqual(
            str(source.url),
            "https://deb.debusine.freexian.com/freexian/debusine-enterprise",
        )
        assert source.signing_key is not None
        self.assertTrue(
            source.signing_key.startswith(
                "-----BEGIN PGP PUBLIC KEY BLOCK-----"
            )
        )

    def test_enterprise(self) -> None:
        source = apt_source(DebusineInstallSource.ENTERPRISE)
        self.assertEqual(source.components, ["main"])
        self.assertEqual(source.suite, "trixie-enterprise")
        self.assertEqual(
            str(source.url),
            "https://deb.debusine.freexian.com/freexian/debusine-enterprise",
        )
        assert source.signing_key is not None
        self.assertTrue(
            source.signing_key.startswith(
                "-----BEGIN PGP PUBLIC KEY BLOCK-----"
            )
        )

    def test_custom(self) -> None:
        custom_source = ExtraExternalRepository(
            url="https://example.net/",
            components=["main"],
            suite="debusine",
            signing_key="PUBLIC KEY",
        )
        source = apt_source(DebusineInstallSource.CUSTOM, custom_source)
        self.assertEqual(source, custom_source)

    def test_custom_source_unspecified(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "debusine_install_source_custom is not specified"
        ):
            apt_source(DebusineInstallSource.CUSTOM)


class CloudInitConfigTests(SimpleTestCase):
    """Tests for CloudInitConfig."""

    def test_full_instantiation(self) -> None:
        config = CloudInitConfig(
            apt=AptConfig(
                preserve_sources_list=False,
                sources_list="deb $MIRROR $RELEASE main",
            ),
            device_aliases={"foo": "/dev/nvme0n1"},
            fs_setup=[
                FilesystemSetup(device="foo", filesystem="ext4"),
            ],
            mounts=[
                ["foo", "/mnt/foo", "ext4", "defaults"],
            ],
            package_reboot_if_required=True,
            package_update=True,
            package_upgrade=True,
            packages=["build-essential", "vim-tiny"],
            bootcmd=["/bin/true"],
            runcmd=["/bin/true", "/bin/false"],
            write_files=[
                WriteFile(
                    path="/etc/hostname", content="test", permissions="0o644"
                ),
            ],
        )
        rendered = config.as_str()
        parsed = yaml.safe_load(rendered)
        self.assertIsInstance(parsed, dict)
        self.assertEqual(
            parsed["apt"],
            {
                "preserve_sources_list": False,
                "sources_list": "deb $MIRROR $RELEASE main",
            },
        )

        self.assertEqual(
            parsed["fs_setup"], [{"device": "foo", "filesystem": "ext4"}]
        )
        self.assertTrue(parsed["package_reboot_if_required"])

    def test_empty_as_str(self) -> None:
        config = CloudInitConfig()
        self.assertEqual(config.as_str(), "#cloud-config\n{}\n")

    def test_simple_as_str(self) -> None:
        config = CloudInitConfig(packages=["foobar"], runcmd=["/bin/true"])
        self.assertEqual(
            config.as_str(),
            dedent(
                """\
                #cloud-config
                packages:
                - foobar
                runcmd:
                - /bin/true
                """
            ),
        )

    def test_simple_as_base64_str(self) -> None:
        config = CloudInitConfig(packages=["foobar"], runcmd=["/bin/true"])
        encoded = config.as_base64_str()
        parsed = yaml.safe_load(b64decode(encoded))
        self.assertEqual(
            parsed, {"packages": ["foobar"], "runcmd": ["/bin/true"]}
        )


class TestWorkerBootstrapCloudInit(TestCase):
    """Tests for worker_bootstrap_cloud_init."""

    def get_file_by_path_if_exists(
        self, files: list[WriteFile], path: str
    ) -> WriteFile | None:
        """Return the file with path in files, if it exists."""
        for file in files:
            if file.path == path:
                return file
        return None

    def get_file_by_path(self, files: list[WriteFile], path: str) -> WriteFile:
        """Return the file with path in files."""
        file = self.get_file_by_path_if_exists(files, path)
        assert file is not None, f"File with {path=} not found in {files=}"
        return file

    def assertFileNotCreated(self, files: list[WriteFile], path: str) -> None:
        """Assert that path is not in files."""
        file = self.get_file_by_path_if_exists(files, path)
        self.assertIsNone(file)

    def test_worker_bootstrap_cloud_init_simple(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.BACKPORTS,
        )
        self.assertTrue(config.package_update)
        self.assertTrue(config.package_upgrade)
        self.assertTrue(config.package_reboot_if_required)
        self.assertEqual(len(config.bootcmd), 2)
        self.assertIn("eatmydata", config.bootcmd[0])
        self.assertEqual("worker", config.hostname)
        self.assertIn("debusine-worker", config.packages)
        self.assertIn("incus", config.packages)
        backports_pin_file = self.get_file_by_path(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )
        self.assertIn("src:debusine", backports_pin_file.content)
        self.assertIn("a=stable-backports", backports_pin_file.content)
        self.assertIn("a=stable\n", backports_pin_file.content)
        self.assertFileNotCreated(
            config.write_files,
            "/etc/apt/preferences.d/90-debusine-ignore-debian",
        )
        token_file = self.get_file_by_path(
            config.write_files, "/etc/debusine/worker/activation-token"
        )
        self.assertEqual(token_file.content, token.key)
        self.assertIsNone(token_file.owner)
        self.get_file_by_path(config.write_files, "/etc/nftables.conf")
        self.assertFileNotCreated(
            config.write_files, "/etc/apt/sources.list.d/debusine.sources"
        )

    def test_worker_bootstrap_cloud_init_daily(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.DAILY_BUILDS,
        )
        self.assertFileNotCreated(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )
        ignore_debian_pin_file = self.get_file_by_path(
            config.write_files,
            "/etc/apt/preferences.d/90-debusine-ignore-debian",
        )
        self.assertIn("release o=Debian", ignore_debian_pin_file.content)
        apt_source = self.get_file_by_path(
            config.write_files, "/etc/apt/sources.list.d/debusine.sources"
        )
        self.assertIn("Suites: trixie-devel", apt_source.content.splitlines())

    def test_worker_bootstrap_cloud_init_enterprise_staging(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.ENTERPRISE_STAGING,
            apt_authentication=BasicAPTAuthenticationData(
                name="enterprise",
                credentials=BasicAPTCredentials(
                    username="test",
                    password="secret-key",
                ),
            ),
        )
        self.assertFileNotCreated(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )
        ignore_debian_pin_file = self.get_file_by_path(
            config.write_files,
            "/etc/apt/preferences.d/90-debusine-ignore-debian",
        )
        self.assertIn("release o=Debian", ignore_debian_pin_file.content)
        apt_source = self.get_file_by_path(
            config.write_files, "/etc/apt/sources.list.d/debusine.sources"
        )
        self.assertIn("Suites: trixie-staging", apt_source.content.splitlines())
        apt_auth = self.get_file_by_path(
            config.write_files, "/etc/apt/auth.conf.d/enterprise.conf"
        )
        self.assertIn("password secret-key", apt_auth.content.splitlines())

    def test_worker_bootstrap_cloud_init_enterprise_no_authentication(
        self,
    ) -> None:
        token = self.playground.create_worker_activation_token()
        with self.assertRaisesRegex(
            ValueError,
            r"apt_authentication is required for ENTERPRISE sources\.",
        ):
            worker_bootstrap_cloud_init(
                "worker",
                token,
                debusine_install_source=DebusineInstallSource.ENTERPRISE_STAGING,
            )

    def test_worker_bootstrap_cloud_init_custom(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.CUSTOM,
            debusine_install_source_custom=ExtraExternalRepository(
                url="https://example.net/",
                components=["main"],
                suite="debusine",
                signing_key="PUBLIC KEY",
            ),
            apt_authentication=BasicAPTAuthenticationData(
                name="examplenet",
                credentials=BasicAPTCredentials(
                    username="test",
                    password="secret-key",
                ),
            ),
        )
        self.assertFileNotCreated(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )
        ignore_debian_pin_file = self.get_file_by_path(
            config.write_files,
            "/etc/apt/preferences.d/90-debusine-ignore-debian",
        )
        self.assertIn("release o=Debian", ignore_debian_pin_file.content)
        apt_source = self.get_file_by_path(
            config.write_files, "/etc/apt/sources.list.d/debusine.sources"
        )
        self.assertIn("Suites: debusine", apt_source.content.splitlines())
        apt_auth = self.get_file_by_path(
            config.write_files, "/etc/apt/auth.conf.d/examplenet.conf"
        )
        self.assertIn(
            "machine https://example.net/", apt_auth.content.splitlines()
        )

    def test_worker_bootstrap_cloud_init_pre_installed(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.PRE_INSTALLED,
        )
        self.assertTrue(config.package_update)
        self.assertTrue(config.package_upgrade)
        self.assertTrue(config.package_reboot_if_required)
        self.assertEqual(config.bootcmd, [])
        self.assertEqual("worker", config.hostname)
        self.assertEqual(
            config.write_files,
            [
                WriteFile(
                    path="/etc/debusine/worker/activation-token",
                    permissions="0o600",
                    content=token.key,
                    owner="debusine-worker:debusine-worker",
                )
            ],
        )
        self.assertEqual(
            config.runcmd, ["systemctl restart debusine-worker.service"]
        )
        self.assertEqual(config.packages, [])

    def test_worker_bootstrap_trixie(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.RELEASE,
            debian_release=DebianRelease.TRIXIE,
        )
        self.assertIsNone(config.apt)
        self.assertFileNotCreated(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )
        self.assertFileNotCreated(
            config.write_files,
            "/etc/apt/preferences.d/90-debusine-ignore-debian",
        )

    def test_worker_bootstrap_trixie_bpo(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = worker_bootstrap_cloud_init(
            "worker",
            token,
            debusine_install_source=DebusineInstallSource.BACKPORTS,
            debian_release=DebianRelease.TRIXIE,
        )
        self.assertIsNone(config.apt)
        self.get_file_by_path(
            config.write_files, "/etc/apt/preferences.d/90-debusine-backports"
        )

    def test_worker_bootstrap_cloud_init_without_key(self) -> None:
        token = Token()
        with self.assertRaisesRegex(
            ValueError,
            (
                r"Requires a fresh token with a key. Token<None> does not have "
                r"one\."
            ),
        ):
            worker_bootstrap_cloud_init(
                "worker",
                token,
                debusine_install_source=DebusineInstallSource.BACKPORTS,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unpack source package and run dpkg-genchanges on it."""

from pathlib import Path
from typing import override

import debusine.tasks.tags as ttags
import debusine.utils
import debusine.worker.tags as wtags
from debusine.artifacts import Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianSourcePackage,
    WorkRequestResults,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
from debusine.tasks.inputs import EnvironmentInput, SingleInput, Stage
from debusine.tasks.models import (
    MakeSourcePackageUploadData,
    MakeSourcePackageUploadDynamicData,
)


class MakeSourcePackageUpload(
    RunCommandTask[
        MakeSourcePackageUploadData, MakeSourcePackageUploadDynamicData
    ],
    BaseTaskWithExecutor[
        MakeSourcePackageUploadData, MakeSourcePackageUploadDynamicData
    ],
):
    """Makes a debian:upload artifact from a debian:source-package artifact."""

    TASK_VERSION = 1

    input_source_artifact = SingleInput(
        field="input.source_artifact",
        categories=[ArtifactCategory.SOURCE_PACKAGE],
        stage=Stage.PENDING,
    )
    environment = EnvironmentInput()

    @override
    def post_init(self) -> None:
        super().post_init()
        self._dsc_path: Path
        self._changes_path: Path
        self._shell_script: Path

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        assert isinstance(self.input_source_artifact.data, DebianSourcePackage)
        provided.add(
            ttags.ProvenanceProvided.SYSTEM,
            {
                ttags.SOURCE_PACKAGE_PREFIX
                + get_source_package_name(self.input_source_artifact.data)
            },
        )

        return provided, required

    @override
    def get_subject(self) -> str | None:
        assert isinstance(self.input_source_artifact.data, DebianSourcePackage)
        return get_source_package_name(self.input_source_artifact.data)

    @override
    def compute_dynamic_data(self) -> MakeSourcePackageUploadDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: source package name of ``input.source_artifact``
        :parameter_summary: source package name and version
        """
        input_source_artifact = self.input_source_artifact
        assert isinstance(input_source_artifact.data, DebianSourcePackage)
        package_name = get_source_package_name(input_source_artifact.data)
        version = get_source_package_version(input_source_artifact.data)
        return MakeSourcePackageUploadDynamicData(
            environment_id=self.environment.artifact_id,
            input_source_artifact_id=input_source_artifact.artifact_id,
            subject=self.get_subject(),
            parameter_summary=f"{package_name}_{version}",
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Populate work directory with user-specified source artifact."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")
        assert self.dynamic_data

        source_artifact_id = self.dynamic_data.input_source_artifact_id
        artifact = self.debusine.artifact_get(
            source_artifact_id, workspace=self.workspace_name
        )
        if artifact.category != ArtifactCategory.SOURCE_PACKAGE:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"input.source_artifact points to a "
                    f"{artifact.category}, not the expected "
                    f"{ArtifactCategory.SOURCE_PACKAGE}."
                ],
            )
            return False

        self.fetch_artifact(source_artifact_id, destination)

        return True

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        r"""
        Find the .dsc file for dpkg-source and dpkg-genchanges.

        Set self._dsc_file to the relevant file.
        Set self._changes_path to the target file.
        Set self._shell_script to a copy of an integration Bash script.

        :param download_directory: where to find the .dsc file
          (downloaded via fetch_input)

        :return: True if valid files were found
        """
        # Find the file to process or early exit if no files
        dsc_files = debusine.utils.find_files_suffixes(
            download_directory, [".dsc"]
        )
        # Ensure we've got 1 .dsc file, see:
        # debusine.artifacts.local_artifacts.SourcePackage.files_contain_one_dsc
        assert len(dsc_files) == 1

        self._dsc_path = dsc_files[0]
        # .changes in the same path as the .dsc for Upload.create()
        self._changes_path = self._dsc_path.with_name(
            self._dsc_path.stem + "_source.changes"
        )

        # Prepare a Bash script: we need to chdir which currently
        # isn't supported (#434), and run multiple commands in _cmdline().
        # Pass files as arguments to avoid shell escaping hell.
        self._shell_script = self._dsc_path.with_name(
            "makesourcepackageupload.sh"
        )
        with open(self._shell_script, "w") as f:
            f.write(
                """#!/bin/bash
download_directory=$(dirname "$1")
cd "$download_directory/"
dpkg-source -x "$1" package/
cd package/
[ -n "$3" ] && optional_args+=("-v$3")
[ -n "$4" ] && optional_args+=("-DDistribution=$4")
dpkg-genchanges -S -sa "${optional_args[@]}" > "$2"
ls -lh "$2"
"""
            )

        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        self.executor_instance.run(
            ["apt-get", "update"], run_as_root=True, check=True
        )
        self.executor_instance.run(
            ["apt-get", "--yes", "install", "dpkg-dev"],
            run_as_root=True,
            check=True,
        )

        return True

    @override
    def _cmdline(self) -> list[str]:
        """Build full makesourcepackageupload.sh command line."""
        return [
            "bash",
            "-x",  # trace commands for debug log
            "-e",  # safely stop on error
            str(self._shell_script),
            str(self._dsc_path),
            str(self._changes_path),
            self.data.since_version or "",
            self.data.target_distribution or "",
        ]

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Create DebianUpload artifact and relationships."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")
        assert self.dynamic_data
        assert self.executor_instance

        self.executor_instance.file_pull(self._changes_path, self._changes_path)

        changes_artifact = Upload.create(
            changes_file=self._changes_path,
        )
        changes_uploaded = self.debusine.upload_artifact(
            changes_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        self.debusine.relation_create(
            changes_uploaded.id,
            self.dynamic_data.input_source_artifact_id,
            RelationType.EXTENDS,
        )
        self.debusine.relation_create(
            changes_uploaded.id,
            self.dynamic_data.input_source_artifact_id,
            RelationType.RELATES_TO,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Decorate 'test' command."""

import sys
from pathlib import Path
from typing import Any, override

from django.conf import settings
from django.core.management import CommandParser
from django.core.management.commands import test

from debusine.django.django_utils import cleanup_temp_directory


class Command(test.Command):
    """Decorate 'test' command."""

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        """Register new --keepdata option, similar to --keepdb."""
        super().add_arguments(parser)
        parser.add_argument(
            "--keepdata",
            action="store_true",
            help="Preserves the test data directory between runs.",
        )

    @override
    def handle(self, *args: Any, **kwargs: Any) -> None:
        """Clean-up or keep temporary test directories hierarchy."""
        try:
            super().handle(*args, **kwargs)
        finally:
            data_path = Path(settings.DEBUSINE_DATA_PATH)
            if kwargs["keepdata"]:
                if kwargs["verbosity"] >= 1:
                    sys.stderr.write(
                        f"Preserving test data directory "
                        f"{str(data_path)!r}...\n"
                    )
            else:
                cleanup_temp_directory()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for signing task models."""

from unittest import TestCase

from debusine.assets.models import KeyPurpose
from debusine.signing.models import SigningMode
from debusine.signing.tasks.models import SignData, SignRepositoryIndexData
from debusine.tasks.models import LookupMultiple


class SignDataTests(TestCase):
    """Tests for Sign task data."""

    def test_purpose_uefi(self) -> None:
        """``purpose=KeyPurpose.UEFI`` is accepted."""
        data = SignData(
            purpose=KeyPurpose.UEFI,
            unsigned=LookupMultiple((1,)),
            key="ABC123",
        )
        self.assertEqual(data.model_dump()["purpose"], KeyPurpose.UEFI)

    def test_purpose_openpgp_upload(self) -> None:
        """``purpose=KeyPurpose.OPENPGP_UPLOAD`` is accepted."""
        data = SignData(
            purpose=KeyPurpose.OPENPGP_UPLOAD,
            unsigned=LookupMultiple((1,)),
            key="ABC123",
        )
        self.assertEqual(
            data.model_dump()["purpose"], KeyPurpose.OPENPGP_UPLOAD
        )

    def test_purpose_openpgp_repository(self) -> None:
        """``purpose=KeyPurpose.OPENPGP_REPOSITORY`` is rejected."""
        with self.assertRaisesRegex(
            ValueError,
            r"Key purpose may not be 'openpgp-repository' here"
            r" \(handled by SignRepositoryIndex task instead\)",
        ):
            SignData(
                purpose=KeyPurpose.OPENPGP_REPOSITORY,
                unsigned=LookupMultiple((1,)),
                key="ABC123",
            )


class SignRepositoryIndexDataTests(TestCase):
    """Tests for SignRepositoryIndex task data."""

    def test_mode_detached(self) -> None:
        """``mode=SigningMode.DETACHED`` is accepted."""
        data = SignRepositoryIndexData(
            suite_collection=1,
            unsigned=2,
            mode=SigningMode.DETACHED,
            signed_name="Release.gpg",
        )
        self.assertEqual(data.model_dump()["mode"], SigningMode.DETACHED)

    def test_mode_clear(self) -> None:
        """``mode=SigningMode.CLEAR`` is accepted."""
        data = SignRepositoryIndexData(
            suite_collection=1,
            unsigned=2,
            mode=SigningMode.CLEAR,
            signed_name="InRelease",
        )
        self.assertEqual(data.model_dump()["mode"], SigningMode.CLEAR)

    def test_mode_not_detached_or_clear(self) -> None:
        """``mode`` must be either ``DETACHED`` or ``CLEAR``."""
        with self.assertRaisesRegex(
            ValueError,
            "Signing mode must be 'detached' or 'clear'; got 'attached'",
        ):
            SignRepositoryIndexData(
                suite_collection=1,
                unsigned=2,
                mode=SigningMode.ATTACHED,
                signed_name="Release.signed",
            )

    def test_signed_name_not_single_path_segment(self) -> None:
        """``signed_name`` must be a single path segment."""
        with self.assertRaisesRegex(
            ValueError, "String should match pattern.*input_value='../naughty'"
        ):
            SignRepositoryIndexData(
                suite_collection=1,
                unsigned=2,
                mode=SigningMode.DETACHED,
                signed_name="../naughty",
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebianPackageBuildLogsManager."""

from typing import override

from django.db.models import IntegerField, Q
from django.db.models.functions import Cast
from django.db.models.lookups import In

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
)
from debusine.client.models import LookupChildType
from debusine.db.models import CollectionItem, default_workspace
from debusine.db.models.lookups import LookupResultArtifact
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections import (
    DebianPackageBuildLogsManager,
    ItemAdditionError,
)
from debusine.server.collections.lookup import lookup_multiple
from debusine.tasks.models import LookupMultiple
from debusine.test.django import TestCase


class DebianPackageBuildLogsManagerTests(TestCase):
    """Tests for DebianPackageBuildLogsManager."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.user = self.playground.get_default_user()
        self.workflow = self.playground.create_worker_task(task_name="noop")
        self.workspace = default_workspace()
        self.collection = self.workspace.get_singleton_collection(
            pc=PermissionContext.for_ui_user(self.user),
            category=CollectionCategory.PACKAGE_BUILD_LOGS,
        )
        self.collection_lookup = f"_@{CollectionCategory.PACKAGE_BUILD_LOGS}"
        self.manager = DebianPackageBuildLogsManager(collection=self.collection)

    def test_do_add_bare_data_no_data(self) -> None:
        """`do_add_bare_data` requires item data."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debian:package-build-logs requires data",
        ):
            self.manager.add_bare_data(
                BareDataCategory.PACKAGE_BUILD_LOG,
                user=self.user,
                workflow=self.workflow,
            )

    def test_do_add_bare_data_raise_item_addition_error(self) -> None:
        """`do_add_bare_data` raises an error for duplicate names."""
        data = {
            "work_request_id": 1,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
        )

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_bare_data(
                BareDataCategory.PACKAGE_BUILD_LOG,
                user=self.user,
                workflow=self.workflow,
                data=data,
            )

    def test_do_add_bare_data_different_work_request_ids(self) -> None:
        """Adding bare data items with different work request IDs is OK."""
        data = {
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }

        items = [
            self.manager.add_bare_data(
                BareDataCategory.PACKAGE_BUILD_LOG,
                user=self.user,
                workflow=self.workflow,
                data={**data, "work_request_id": work_request_id},
            )
            for work_request_id in (1, 2)
        ]

        self.assertEqual(
            [item.name for item in items],
            [
                "debian_bookworm_amd64_hello_1.0-1_1",
                "debian_bookworm_amd64_hello_1.0-1_2",
            ],
        )

    def test_do_add_bare_data_replace(self) -> None:
        """`do_add_bare_data` can replace an existing bare data item."""
        data = {
            "work_request_id": 1,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        item_old = self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
        )

        item_new = self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )

        item_old.refresh_from_db()
        self.assertEqual(item_old.name, "debian_bookworm_amd64_hello_1.0-1_1")
        self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_old.data, data)
        self.assertEqual(item_old.removed_by_user, self.user)
        self.assertEqual(item_old.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(item_new.name, "debian_bookworm_amd64_hello_1.0-1_1")
        self.assertEqual(item_new.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_new.data, data)
        self.assertIsNone(item_new.removed_at)

    def test_do_add_bare_data_replace_nonexistent(self) -> None:
        """Replacing a nonexistent bare data item is allowed."""
        data = {
            "work_request_id": 1,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }

        item = self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )

        self.assertEqual(item.name, "debian_bookworm_amd64_hello_1.0-1_1")
        self.assertEqual(item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item.data, data)

    def test_do_add_artifact_no_variables(self) -> None:
        """`do_add_artifact` requires variables."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
        )

        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debian:package-build-logs requires variables",
        ):
            self.manager.add_artifact(
                artifact, user=self.user, workflow=self.workflow
            )

    def test_do_add_artifact_raise_item_addition_error(self) -> None:
        """`do_add_artifact` raises an error for duplicate names."""
        work_request = self.playground.create_worker_task()
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )
        artifact_2, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
        }
        self.manager.add_artifact(
            artifact_1, user=self.user, workflow=self.workflow, variables=data
        )

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_artifact(
                artifact_2,
                user=self.user,
                workflow=self.workflow,
                variables=data,
            )

    def test_do_add_artifact_override_srcpkg_name_version(self) -> None:
        """
        `do_add_artifact` can override the source package name/version.

        This isn't very useful in practice, but it allows its interface to
        be more like `do_add_bare_data`.
        """
        work_request = self.playground.create_worker_task()
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello-2",
            "srcpkg_version": "1.0-2",
        }

        item = self.manager.add_artifact(
            artifact, user=self.user, workflow=self.workflow, variables=data
        )

        self.assertEqual(
            item.name, f"debian_bookworm_amd64_hello-2_1.0-2_{work_request.id}"
        )
        self.assertEqual(item.data, data)

    def test_do_add_artifact_different_work_request_ids(self) -> None:
        """Adding artifacts with different work request IDs is OK."""
        work_request_1 = self.playground.create_worker_task()
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )
        work_request_2 = self.playground.create_worker_task()
        artifact_2, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )
        data = {
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
        }

        items = [
            self.manager.add_artifact(
                artifact_1,
                user=self.user,
                workflow=self.workflow,
                variables={**data, "work_request_id": work_request.id},
            )
            for artifact, work_request in (
                (artifact_1, work_request_1),
                (artifact_2, work_request_2),
            )
        ]

        self.assertEqual(
            [item.name for item in items],
            [
                f"debian_bookworm_amd64_hello_1.0-1_{work_request_1.id}",
                f"debian_bookworm_amd64_hello_1.0-1_{work_request_2.id}",
            ],
        )

    def test_do_add_artifact_replace_bare_data(self) -> None:
        """`do_add_artifact` can replace an existing bare data item."""
        work_request = self.playground.create_worker_task()
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        item_old = self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
        )
        artifact_new, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data={"source": "hello", "version": "1.0-1"},
        )

        item_new = self.manager.add_artifact(
            artifact_new,
            user=self.user,
            workflow=self.workflow,
            variables=data,
            replace=True,
        )

        item_old.refresh_from_db()
        self.assertEqual(
            item_old.name,
            f"debian_bookworm_amd64_hello_1.0-1_{work_request.id}",
        )
        self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_old.data, data)
        self.assertEqual(item_old.removed_by_user, self.user)
        self.assertEqual(item_old.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(
            item_new.name,
            f"debian_bookworm_amd64_hello_1.0-1_{work_request.id}",
        )
        self.assertEqual(item_new.artifact, artifact_new)
        self.assertEqual(item_new.data, data)
        self.assertIsNone(item_new.removed_at)

    def test_do_add_artifact_replace_artifact(self) -> None:
        """`do_add_artifact` can replace an existing artifact."""
        worker_1 = self.playground.create_worker()
        worker_2 = self.playground.create_worker()
        work_request = self.playground.create_worker_task(worker=worker_1)
        artifact_old, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
        )
        artifact_new, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
        )
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        item_old = self.manager.add_artifact(
            artifact_old, user=self.user, workflow=self.workflow, variables=data
        )

        work_request.assign_worker(worker_2)
        item_new = self.manager.add_artifact(
            artifact_new,
            user=self.user,
            workflow=self.workflow,
            variables=data,
            replace=True,
        )

        item_old.refresh_from_db()
        self.assertEqual(
            item_old.name,
            f"debian_bookworm_amd64_hello_1.0-1_{work_request.id}",
        )
        self.assertEqual(item_old.artifact, artifact_old)
        self.assertEqual(item_old.data, {**data, "worker": worker_1.name})
        self.assertEqual(item_old.removed_by_user, self.user)
        self.assertEqual(item_old.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(
            item_new.name,
            f"debian_bookworm_amd64_hello_1.0-1_{work_request.id}",
        )
        self.assertEqual(item_new.artifact, artifact_new)
        self.assertEqual(item_new.data, {**data, "worker": worker_2.name})
        self.assertIsNone(item_new.removed_at)

    def test_do_add_artifact_replace_nonexistent(self) -> None:
        """Replacing a nonexistent artifact is allowed."""
        work_request = self.playground.create_worker_task()
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
        )
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }

        item = self.manager.add_artifact(
            artifact,
            user=self.user,
            workflow=self.workflow,
            variables=data,
            replace=True,
        )

        self.assertEqual(
            item.name, f"debian_bookworm_amd64_hello_1.0-1_{work_request.id}"
        )
        self.assertEqual(item.artifact, artifact)
        self.assertEqual(item.data, data)

    def test_do_remove_item_bare_data(self) -> None:
        """`do_remove_item` removes a bare data item."""
        data = {
            "work_request_id": 1,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        item = self.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.user,
            workflow=self.workflow,
            data=data,
            replace=True,
        )

        self.manager.remove_item(item, user=self.user, workflow=self.workflow)

        self.assertEqual(item.removed_by_user, self.user)
        self.assertEqual(item.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item.removed_at)

    def test_do_remove_item_artifact(self) -> None:
        """`do_remove_item` removes an artifact item."""
        work_request = self.playground.create_worker_task()
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.PACKAGE_BUILD_LOG, data={}
        )
        data = {
            "work_request_id": work_request.id,
            "vendor": "debian",
            "codename": "bookworm",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        item = self.manager.add_artifact(
            artifact,
            user=self.user,
            workflow=self.workflow,
            variables=data,
            replace=True,
        )

        self.manager.remove_item(item, user=self.user, workflow=self.workflow)

        self.assertEqual(item.removed_by_user, self.user)
        self.assertEqual(item.removed_by_workflow, self.workflow)
        self.assertIsNotNone(item.removed_at)

    def test_do_lookup_filter_unexpected_format(self) -> None:
        """`do_lookup_filter` raises LookupError for an unexpected format."""
        with self.assertRaisesRegex(
            LookupError, r'^Unexpected lookup filter format: "invalid"'
        ):
            self.manager.lookup_filter(
                "invalid",
                "foo",
                workspace=self.workspace,
                pc=PermissionContext.for_ui_user(self.user),
            )

    def test_do_lookup_filter_same_work_request_single(self) -> None:
        """`do_lookup_filter`: `same_work_request` with a single lookup."""
        work_requests = [self.playground.create_worker_task() for _ in range(2)]
        upload, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD, work_request=work_requests[0]
        )
        build_logs = [
            self.playground.create_artifact(
                category=ArtifactCategory.PACKAGE_BUILD_LOG,
                work_request=work_request,
            )[0]
            for work_request in work_requests
        ]
        items = [
            self.manager.add_artifact(
                build_log,
                user=self.user,
                variables={
                    "work_request_id": build_log.created_by_work_request_id,
                    "vendor": "debian",
                    "codename": "bookworm",
                    "architecture": "amd64",
                    "srcpkg_name": "hello",
                    "srcpkg_version": "1.0-1",
                },
            )
            for build_log in build_logs
        ]
        subordinate_lookup = f"{upload.id}@artifacts"

        condition = self.manager.lookup_filter(
            "same_work_request",
            subordinate_lookup,
            workspace=self.workspace,
            pc=PermissionContext.for_ui_user(self.user),
        )
        self.assertEqual(
            condition,
            Q(
                In(
                    Cast("data__work_request_id", IntegerField()),
                    {work_requests[0].id},
                )
            ),
        )
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition), [items[0]]
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "lookup__same_work_request": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.workspace,
                pc=PermissionContext.for_ui_user(self.user),
                expect_type=LookupChildType.ARTIFACT,
            ),
            [
                LookupResultArtifact(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=items[0],
                    artifact=build_logs[0],
                )
            ],
        )

    def test_do_lookup_filter_same_work_request_multiple(self) -> None:
        """`do_lookup_filter`: `same_work_request` with a multiple lookup."""
        work_requests = [self.playground.create_worker_task() for _ in range(3)]
        uploads = [
            self.playground.create_artifact(
                category=ArtifactCategory.UPLOAD, work_request=work_request
            )[0]
            for work_request in work_requests[:2]
        ]
        build_logs = [
            self.playground.create_artifact(
                category=ArtifactCategory.PACKAGE_BUILD_LOG,
                work_request=work_request,
            )[0]
            for work_request in work_requests
        ]
        items = [
            self.manager.add_artifact(
                build_log,
                user=self.user,
                variables={
                    "work_request_id": build_log.created_by_work_request_id,
                    "vendor": "debian",
                    "codename": "bookworm",
                    "architecture": "amd64",
                    "srcpkg_name": "hello",
                    "srcpkg_version": "1.0-1",
                },
            )
            for build_log in build_logs
        ]
        subordinate_lookup = LookupMultiple.model_validate(
            [f"{upload.id}@artifacts" for upload in uploads]
        )

        condition = self.manager.lookup_filter(
            "same_work_request",
            subordinate_lookup,
            workspace=self.workspace,
            pc=PermissionContext.for_ui_user(self.user),
        )
        self.assertEqual(
            condition,
            Q(
                In(
                    Cast("data__work_request_id", IntegerField()),
                    {work_request.id for work_request in work_requests[:2]},
                )
            ),
        )
        self.assertQuerySetEqual(
            self.collection.child_items.active().filter(condition),
            items[:2],
            ordered=False,
        )
        lookup = LookupMultiple.model_validate(
            {
                "collection": self.collection_lookup,
                "lookup__same_work_request": subordinate_lookup,
            }
        )
        self.assertCountEqual(
            lookup_multiple(
                lookup,
                self.workspace,
                pc=PermissionContext.for_ui_user(self.user),
                expect_type=LookupChildType.ARTIFACT,
            ),
            [
                LookupResultArtifact(
                    lookup=lookup.root[0],
                    parent_collection_lookup=self.collection_lookup,
                    collection_item=item,
                    artifact=build_log,
                )
                for item, build_log in zip(items[:2], build_logs[:2])
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Catalog of known worker tags and worker tag prefixes."""

from enum import StrEnum

from debusine.utils.tag_utils import CandidateTags, ProvenanceRestrictions


class ProvenanceProvided(StrEnum):
    """Provenances for worker-provided worker tags."""

    #: Tags provided in dynamic worker metadata when a worker registers as
    #: active
    WORKER = "worker"
    #: Tags provided in static worker metadata by the Debusine admins
    ADMIN = "admin"
    #: Tags from database queries
    SYSTEM = "system"


class ProvenanceRequired(StrEnum):
    """Provenances for task-required worker tags."""

    #: Tags from database queries
    SYSTEM = "system"
    #: Tags from workspace task configuration
    WORKSPACE = "workspace"


class TaskRequiredTags(CandidateTags[ProvenanceRequired]):
    """Candidate task-required tags."""


#: Prefix for worker build architecture
BUILD_ARCH_PREFIX = "worker:build-arch:"

#: Prefix for worker capability tags
CAP_PREFIX = "worker:cap:"

#: Prefix for worker classes
CLASS_PREFIX = "worker:class:"

#: Prefix and tags for worker type tags
WORKER_TYPE_PREFIX = "worker:type:"
WORKER_TYPE_CELERY = WORKER_TYPE_PREFIX + "celery"
WORKER_TYPE_EXTERNAL = WORKER_TYPE_PREFIX + "external"
WORKER_TYPE_SIGNING = WORKER_TYPE_PREFIX + "signing"
WORKER_TYPE_NOT_ASSIGNABLE = WORKER_TYPE_PREFIX + "not-assignable"

#: Prefix for worker executor tags
EXECUTOR_PREFIX = "worker:executor:"

#: Prefix for worker task availability tags
TASK_PREFIX = "worker:task:"

#: Restrictions applied when adding provided tags
restrictions_when_provided = ProvenanceRestrictions("worker-provided")
# worker:class:* tags are only allowed to be provided by admins
restrictions_when_provided.add_prefix(CLASS_PREFIX, [ProvenanceProvided.ADMIN])
# worker:type:* tags are only allowed to come from database queries
restrictions_when_provided.add_prefix(
    WORKER_TYPE_PREFIX, [ProvenanceProvided.SYSTEM]
)

#: Restrictions applied when adding required tags
# Use empty restrictions, since adding required tags is not a security
# sensitive operation, but removing them is
restrictions_when_required = ProvenanceRestrictions("worker-required")
